/*********************************************************************************************************************************
 * Copyright (c) Des Herriott 1993 - 1994
 *               Erik Kunze   1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: Des Herriott
 *          Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: sna.c,v 1.1 2000/07/13 15:03:33 erik Rel $";
#endif
#include <stdio.h>
#include <string.h>
#include "cpu.h"
#include "debug.h"
#include "resource.h"
#include "mem.h"
#include "io.h"
#include "machine.h"
#include "util.h"
#include "menu.h"
#include "sna.h"
typedef struct _snaHeader {
	uns8 i;
	uns8 l_alt, h_alt;
	uns8 e_alt, d_alt;
	uns8 c_alt, b_alt;
	uns8 f_alt, a_alt;
	uns8 l, h;
	uns8 e, d;
	uns8 c, b;
	uns8 liy, hiy;
	uns8 lix, hix;
	uns8 iff2;
	uns8 r;
	uns8 f, a;
	uns8 lsp, hsp;
	uns8 im;
	uns8 border;
} snaHeader;
#ifdef DEBUG
#define DEB(x)							{ if ((CFGGET(debug) & D_SNAP) != 0) { x } }
#else
#define DEB(x)
#endif
int
SnaRead(char *fname)
{
	int readonly;
	size_t size;
	void *mmp;
	snaHeader *header;
	int ret = -1;
	readonly = 1;
	mmp = Mmap(fname, 1, &readonly, &size);
	if (mmp == (void *)-1)
	{
		goto quit;
	}
	if (size != 49179)
	{
		Msg(M_ERR, "snapshot <%s> is not in 48K SNA format", fname);
		goto quit;
	}
	Msg(M_INFO, "snapshot format is SNA");
	Msg(M_INFO, "hardware type is 48K");
	header = (snaHeader *)mmp;
#ifdef XZX_IF1
	if (CFGGET(if1) > 0)
	{
		CFGSET(if1, 0);
		MenuFileCartridgeUpdate();
	}
#endif
	MachineSelect(SP_48_3, 0);
	I    = header->i;
	L1   = header->l_alt;
	H1   = header->h_alt;
	E1   = header->e_alt;
	D1   = header->d_alt;
	C1   = header->c_alt;
	B1   = header->b_alt;
	F1   = header->f_alt;
	A1   = header->a_alt;
	L    = header->l;
	H    = header->h;
	E    = header->e;
	D    = header->d;
	C    = header->c;
	B    = header->b;
	LIY  = header->liy;
	HIY  = header->hiy;
	LIX  = header->lix;
	HIX  = header->hix;
	IFF2 = (header->iff2 >> 2) & 0x01;
	R    = header->r;
	F    = header->f;
	A    = header->a;
	LSP  = header->lsp;
	HSP  = header->hsp;
	IM   = header->im;
	Machine->outPort(P_ULA, header->border);
	(void)memcpy(RealMemory[RAM5], (uns8 *)mmp + sizeof(snaHeader), XZX_PAGESIZE);
	(void)memcpy(RealMemory[RAM2], (uns8 *)mmp + sizeof(snaHeader) + XZX_PAGESIZE, XZX_PAGESIZE);
	(void)memcpy(RealMemory[RAM0], (uns8 *)mmp + sizeof(snaHeader) + XZX_PAGESIZE * 2, XZX_PAGESIZE);
	RETN();
	ret = 0;
  quit:
	if (mmp != (void *)-1)
	{
		Munmap(mmp, size);
	}
	return ret;
}
int
SnaWrite(char *fname)
{
	FILE *fp;
	snaHeader header;
	int ret = -1;
	PUSH(HPC, LPC);
	if ((fp = Fopen(fname, "wb")) == (FILE *)0)
	{
		Msg(M_PERR, "couldn't open <%s> for writing", fname);
		goto quit;
	}
	header.i     = I;
	header.l_alt = L1;
	header.h_alt = H1;
	header.e_alt = E1;
	header.d_alt = D1;
	header.c_alt = C1;
	header.b_alt = B1;
	header.f_alt = F1;
	header.a_alt = A1;
	header.l     = L;
	header.h     = H;
	header.c     = C;
	header.b     = B;
	header.liy   = LIY;
	header.hiy   = HIY;
	header.lix   = LIX;
	header.hix   = HIX;
	header.iff2  = IFF2 << 2;
	header.r     = R;
	header.f     = F;
	header.a     = A;
	header.lsp   = LSP;
	header.hsp   = HSP;
	header.im    = IM;
	header.border = Last0xFE & B_BORDER;
	if (fwrite(&header, sizeof(header), 1, fp) != 1)
	{
		Msg(M_PERR, "couldn't write snapshot header");
		goto quit;
	}
	if (fwrite(RealMemory[RAM5], XZX_PAGESIZE, 1, fp) != 1
		|| fwrite(RealMemory[RAM2], XZX_PAGESIZE, 1, fp) != 1
		|| fwrite(RealMemory[RAM0], XZX_PAGESIZE, 1, fp) != 1)
	{
		Msg(M_PERR, "couldn't write memory dump");
		goto quit;
	}
	ret = 0;
  quit:
	if (fp != (FILE *)0)
	{
		(void)fclose(fp);
	}
	POP(HPC, LPC);
	return ret;
}
