/********************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation
 * for any purpose is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that the name
 * of the copyright holder not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.  The
 * copyright holder makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 *******************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: fdc.c,v 4.33 2000/11/07 05:15:47 erik Rel $";
#endif
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "debug.h"
#include "resource.h"
#include "io.h"
#include "util.h"
#include "menu.h"
#include "sbar.h"
#include "ui.h"
#include "fdc.h"
#define STM_RDY			0x80
#define STM_DIR			0x40
#define STM_EXEC		0x20
#define STM_BUSY		0x10
#define STM_BUSY3		0x08
#define STM_BUSY2		0x04
#define STM_BUSY1		0x02
#define STM_BUSY0		0x01
#define ST0_ERRCODE		0xc0
#define ST0_OK			0x00
#define ST0_NOTFINISH	0x40
#define ST0_INVALCMD	0x80
#define ST0_STOPPED		0xc0
#define ST0_SEEKEND		0x20
#define ST0_EQFAULT		0x10
#define ST0_NOTRDY		0x08
#define ST0_SIDE		0x04
#define ST0_DRIVE		0x03
#define ST1_ENDCYL		0x80
#define ST1_DATAERR		0x20
#define ST1_OVERRUN		0x10
#define ST1_NODATA		0x04
#define ST1_WRPROT		0x02
#define ST1_ADDRMARK	0x01
#define ST2_CTRLMARK	0x40
#define ST2_DATAERR		0x20
#define ST2_ERRCYLTRK	0x10
#define ST2_SCANOK		0x08
#define ST2_ERRSCAN		0x04
#define ST2_BADCYL		0x02
#define ST2_ADDRMARK	0x01
#define ST3_FAULT		0x80
#define ST3_WRPROT		0x40
#define ST3_RDY			0x20
#define ST3_TRK0		0x10
#define ST3_DS			0x08
#define ST3_SIDE		0x04
#define ST3_DRIVE		0x03
typedef struct {
	uns8 track;
	uns8 head;
	uns8 sector;
	uns8 bps;
	uns8 st1;
	uns8 st2;
	uns8 unused[2];
} sectorInfo;
typedef struct {
	uns8 magic[13];
	uns8 unused[3];
	uns8 track;
	uns8 head;
	uns8 unused2[2];
	uns8 bps;
	uns8 spt;
	uns8 gap3;
	uns8 fill;
	sectorInfo si[29];
} trackInfo;
typedef struct {
	uns8 magic[34];
	uns8 creator[14];
	uns8 tracks;
	uns8 sides;
	uns8 tracksize[2];
	uns8 unused[204];
} diskInfo;
typedef struct {
	uns8 tracks;
	uns8 sides;
	uns16 tracksize;
	char *fn;
	int readonly;
	uns8 *mmp;
	size_t mmsize;
} diskDef;
typedef struct {
	unsigned int id;
	unsigned int sides;
	unsigned int tracks;
	diskDef disk;
	unsigned int side;
	unsigned int track;
	trackInfo *trackp;
	unsigned int sector;
	uns8 *sectorp;
} driveDef;
#define DRIVE			(activeDrive)
#define DISK			((diskDef *)(&(activeDrive->disk)))
#define RESET_FDC_STATE	{ FdcFunction = fdcStartCommand; FdcStm = STM_RDY; }
#define MOTOR_ON		(Last0x1FFD & B_DISK)
#define DRIVE_READY		(MOTOR_ON && DISK->fn)
#define DRIVE_LED_ON	SbarSetLed(LED_ON);
#define DRIVE_LED_OFF	SbarSetLed(LED_OFF);
#ifdef DEBUG
#define DEB(x)			{ if (CFGGET(debug) & D_UPD765A) { x } }
#else
#define DEB(x)
#endif
static uns8 fdcStartCommand(uns8);
static uns8 fdcReadData(uns8);
static uns8 fdcReadDeletedData(uns8);
static uns8 fdcWriteData(uns8);
static uns8 fdcWriteDeletedData(uns8);
static uns8 fdcReadTrack(uns8);
static uns8 fdcReadSectorId(uns8);
static uns8 fdcFormatTrack(uns8);
static uns8 fdcScanEqual(uns8);
static uns8 fdcScanLowOrEqual(uns8);
static uns8 fdcScanHighOrEqual(uns8);
static uns8 fdcRecalibrate(uns8);
static uns8 fdcSenseIntStatus(uns8);
static uns8 fdcSpecify(uns8);
static uns8 fdcSenseDriveStatus(uns8);
static uns8 fdcSeek(uns8);
static uns8 fdcError(uns8);
static void drvSelect(uns8, int);
static void drvSeek(unsigned int);
static void drvNextSector(void);
static sectorInfo *drvFindSector(void);
static int diskOpenImage(int, char *);
static void diskCloseImage(int);
uns8 (*FdcFunction)(uns8) = fdcStartCommand;
uns8 FdcStm;
static uns8 st0, st1, st2, st3;
static int intPending;
static int cntBytes;
static driveDef drives[4], *activeDrive;
static sectorInfo si;
static int skip;
static int multi;
static int deleted;
static uns8 last, gap, length;
static int seclen;
void
FdcInit(void)
{
	int i;
	DRIVE = &drives[0];
	for (i = 0; i < 2; i++)
	{
		drives[i].id = i;
		drives[i].sides = 2;
		drives[i].tracks = 80;
		drives[i].trackp = (trackInfo *)-1;
		drives[i].sectorp = (uns8 *)-1;
		if (diskOpenImage(i, CFGGET(sp3Disks[i])) < 0)
		{
			free((void *)CFGGET(sp3Disks[i]));
		}
		CFGSET(sp3Disks[i], (char *)0);
	}
}
void
FdcReset(void)
{
	st0 = 0;
	st1 = 0;
	st2 = 0;
	st3 = 0;
	RESET_FDC_STATE;
	intPending = 0;
	drvSelect(0, 0);
}
int
FdcInsertDisk(int drive)
{
	int types[2] = { FT_DSK, FT_END };
	char *filename;
	int ret = -1;
	if ((filename = UiFileSelector("Insert Disk", 1, types)))
	{
		if (!diskOpenImage(drive, filename))
		{
			ret = 0;
		}
		else
		{
			free(filename);
		}
		MenuFileDiskUpdate();
	}
	return ret;
}
void
FdcEjectDisk(int drive)
{
	diskCloseImage(drive);
	MenuFileDiskUpdate();
}
char *
FdcDiskName(int drive)
{
	return (drives[drive].disk.fn);
}
static uns8
fdcStartCommand(uns8 val)
{
	static uns8 (*cmds[32])(uns8) = {
		fdcError,
		fdcError,
		fdcReadTrack,
		fdcSpecify,
		fdcSenseDriveStatus,
		fdcWriteData,
		fdcReadData,
		fdcRecalibrate,
		fdcSenseIntStatus,
		fdcWriteDeletedData,
		fdcReadSectorId,
		fdcError,
		fdcReadDeletedData,
		fdcFormatTrack,
		fdcError,
		fdcSeek,
		fdcError,
		fdcScanEqual,
		fdcError,
		fdcError,
		fdcError,
		fdcError,
		fdcError,
		fdcError,
		fdcError,
		fdcScanLowOrEqual,
		fdcError,
		fdcError,
		fdcError,
		fdcScanHighOrEqual,
		fdcError,
		fdcError,
	};
	cntBytes = 0;
	FdcFunction = cmds[val & 0x1f];
	return (FdcFunction(val));
}
static uns8
fdcReadData(uns8 val)
{
	sectorInfo *secInf;
	switch (cntBytes++)
	{
		case 0:
			deleted = 0;
			skip = (val >> 5) & 0x01;
			multi = val >> 7;
			DEB(Msg(M_DEBUG, "FDC: READ DATA");
				Msg(M_DEBUG, "      MT %d, MFM %d, SK %d",
					multi, (val >> 6) & 0x01, skip););
			FdcStm |= STM_BUSY;
			break;
		case 1:
			drvSelect(val, 1);
			DEB(Msg(M_DEBUG, "      S %d, DS %d", DRIVE->side, DRIVE->id););
			break;
		case 2:
			si.track = val;
			DEB(Msg(M_DEBUG, "      C %d", val););
			break;
		case 3:
			si.head = val;
			DEB(Msg(M_DEBUG, "      H %d", val););
			break;
		case 4:
			si.sector = val;
			DEB(Msg(M_DEBUG, "      R %d", val););
			break;
		case 5:
			si.bps = val;
			DEB(Msg(M_DEBUG, "      N %d", val););
			break;
		case 6:
			last = val;
			DEB(Msg(M_DEBUG, "      EOT %d", val););
			break;
		case 7:
			gap = val;
			DEB(Msg(M_DEBUG, "      GPL %d", val););
			break;
		case 8:
			length = val;
			DEB(Msg(M_DEBUG, "      DTL %d", val););
			st0 = DRIVE->id | (DRIVE->side << 2);
			st1 = 0;
			st2 = 0;
			FdcStm |= STM_DIR;
			if (!DRIVE_READY)
			{
				st0 |= ST0_STOPPED | ST0_NOTRDY;
			}
			else if (DRIVE->trackp == (trackInfo *)-1
					 || DRIVE->sectorp == (uns8 *)-1)
			{
				st0 |= ST0_NOTFINISH;
				st1 |= ST1_ADDRMARK;
			}
			else
			{
			  next:
				if (!(secInf = drvFindSector()))
				{
					st0 |= ST0_NOTFINISH;
					st1 |= ST1_NODATA;
					if (DRIVE->trackp->si[DRIVE->sector].track != si.track)
					{
						st2 |= si.track == 0xff ? ST2_BADCYL : ST2_ERRCYLTRK;
					}
					break;
				}
				st1 = secInf->st1;
				st2 = secInf->st2;
				if (deleted)
				{
					st2 ^= ST2_CTRLMARK;
				}
				if (skip && (st2 & ST2_CTRLMARK))
				{
				  check:
					if (si.sector != last)
					{
						si.sector++;
						goto next;
					}
					else
					{
						st0 |= ST0_NOTFINISH;
						st1 |= ST1_ENDCYL;
					}
				}
				else
				{
					seclen = 128 << secInf->bps;
					cntBytes = -(si.bps ? seclen : length > 128 ? 128 : length);
					FdcStm |= STM_EXEC;
				}
			}
			break;
		default:
			val = seclen-- > 0 ? *DRIVE->sectorp++ : (uns8)rand();
			if (cntBytes == 0)
			{
				cntBytes = 9;
				FdcStm &= ~STM_EXEC;
				if (!(st2 & ST2_CTRLMARK))
				{
					goto check;
				}
			}
			break;
		case 9:
			val = st0;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 10:
			val = st1;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 11:
			val = st2;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 12:
			val = si.track;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 13:
			val = si.head;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 14:
			val = si.sector;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 15:
			val = si.bps;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			intPending = 1;
			RESET_FDC_STATE;
			break;
	}
	return val;
}
static uns8
fdcReadDeletedData(uns8 val)
{
	cntBytes++;
	deleted = 1;
	skip = (val >> 5) & 0x01;
	multi = val >> 7;
	DEB(Msg(M_DEBUG, "FDC: READ DELETED DATA");
		Msg(M_DEBUG, "      MT %d, MFM %d, SK %d",
			multi, (val >> 6) & 0x01, skip););
	FdcFunction = fdcReadData;
	FdcStm |= STM_BUSY;
	return val;
}
static uns8
fdcWriteData(uns8 val)
{
	sectorInfo *secInf;
	switch (cntBytes++)
	{
		case 0:
			multi = val >> 7;
			DEB(Msg(M_DEBUG, "FDC: WRITE DATA");
				Msg(M_DEBUG, "      MT %d, MFM %d",
					multi, (val >> 6) & 0x01););
			FdcStm |= STM_BUSY;
			break;
		case 1:
			drvSelect(val, 1);
			DEB(Msg(M_DEBUG, "      S %d, DS %d", DRIVE->side, DRIVE->id););
			break;
		case 2:
			si.track = val;
			DEB(Msg(M_DEBUG, "      C %d", val););
			break;
		case 3:
			si.head = val;
			DEB(Msg(M_DEBUG, "      H %d", val););
			break;
		case 4:
			si.sector = val;
			DEB(Msg(M_DEBUG, "      R %d", val););
			break;
		case 5:
			si.bps = val;
			DEB(Msg(M_DEBUG, "      N %d", val););
			break;
		case 6:
			last = val;
			DEB(Msg(M_DEBUG, "      EOT %d", val););
			break;
		case 7:
			gap = val;
			DEB(Msg(M_DEBUG, "      GPL %d", val););
			break;
		case 8:
			length = val;
			DEB(Msg(M_DEBUG, "      DTL %d", val););
			st0 = DRIVE->id | (DRIVE->side << 2);
			st1 = 0;
			st2 = 0;
			FdcStm |= STM_DIR;
			if (!DRIVE_READY)
			{
				st0 |= ST0_STOPPED | ST0_NOTRDY;
			}
			else if (DRIVE->trackp == (trackInfo *)-1
					 || DRIVE->sectorp == (uns8 *)-1)
			{
				st0 |= ST0_NOTFINISH;
				st1 |= ST1_ADDRMARK;
			}
			else if (DISK->readonly)
			{
				st0 |= ST0_NOTFINISH;
				st1 |= ST1_WRPROT;
			}
			else
			{
			  next:
				if ((secInf = drvFindSector()))
				{
					st1 = secInf->st1;
					st2 = (secInf->st2 |= deleted ? ST2_CTRLMARK : 0);
					seclen = 128 << secInf->bps;
					cntBytes = -(si.bps ? seclen : length > 128 ? 128 : length);
					FdcStm &= ~STM_DIR;
					FdcStm |= STM_EXEC;
				}
				else
				{
					st0 |= ST0_NOTFINISH;
					st1 |= ST1_ENDCYL | ST1_NODATA | ST1_ADDRMARK;
				}
			}
			break;
		default:
			if (seclen--)
			{
				*DRIVE->sectorp++ = val;
			}
			if (cntBytes == 0)
			{
				cntBytes = 9;
				FdcStm &= ~STM_EXEC;
				if (si.sector != last)
				{
					si.sector++;
					goto next;
				}
				st1 |= ST1_ENDCYL;
				FdcStm |= STM_DIR;
			}
			break;
		case 9:
			val = st0;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 10:
			val = st1;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 11:
			val = st2;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 12:
			val = si.track;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 13:
			val = si.head;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 14:
			val = si.sector;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 15:
			val = si.bps;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			intPending = 1;
			RESET_FDC_STATE;
			break;
	}
	return val;
}
static uns8
fdcWriteDeletedData(uns8 val)
{
	cntBytes++;
	deleted = 1;
	multi = val >> 7;
	DEB(Msg(M_DEBUG, "FDC: WRITE DELETED DATA");
		Msg(M_DEBUG, "      MT %d, MFM %d", multi, (val >> 6) & 0x01););
	FdcFunction = fdcWriteData;
	FdcStm |= STM_BUSY;
	return val;
}
static uns8
fdcReadTrack(uns8 val)
{
	Msg(M_WARN, "FDC: READ TRACK");
	return fdcError(val);
}
static uns8
fdcReadSectorId(uns8 val)
{
	switch (cntBytes++)
	{
		case 0:
			DEB(Msg(M_DEBUG, "FDC: READ SECTOR ID");
				Msg(M_DEBUG, "      MFM %d", (val >> 6) & 0x01););
			FdcStm |= STM_BUSY;
			break;
		case 1:
			drvSelect(val, 1);
			DEB(Msg(M_DEBUG, "      S %d, DS %d", DRIVE->side, DRIVE->id););
			st0 = DRIVE->id | (DRIVE->side << 2);
			st1 = 0;
			st2 = 0;
			FdcStm |= STM_DIR;
			if (DRIVE_READY)
			{
				if (DRIVE->trackp == (trackInfo *)-1
					|| DRIVE->sectorp == (uns8 *)-1)
				{
					st0 |= ST0_NOTFINISH;
					st1 |= ST1_ADDRMARK;
					(void)memset(&si, 0xff, sizeof(sectorInfo));
				}
				else
				{
					drvNextSector();
					(void)memcpy(&si, &(DRIVE->trackp->si[DRIVE->sector]),
								 sizeof(sectorInfo));
				}
			}
			else
			{
				st0 |= ST0_STOPPED | ST0_NOTRDY;
			}
			st0 |= ST0_SEEKEND;
			break;
		case 2:
			val = st0;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 3:
			val = st1;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 4:
			val = st2;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 5:
			val = si.track;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 6:
			val = si.head;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 7:
			val = si.sector;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 8:
			val = si.bps;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			intPending = 1;
			RESET_FDC_STATE;
			break;
	}
	return val;
}
static uns8
fdcFormatTrack(uns8 val)
{
	Msg(M_WARN, "FDC: FORMAT TRACK");
	return fdcError(val);
}
static uns8
fdcScanEqual(uns8 val)
{
	Msg(M_WARN, "FDC: SCAN EQUAL");
	return fdcError(val);
}
static uns8
fdcScanLowOrEqual(uns8 val)
{
	Msg(M_WARN, "FDC: SCAN LOW OR EQUAL");
	return fdcError(val);
}
static uns8
fdcScanHighOrEqual(uns8 val)
{
	Msg(M_WARN, "FDC: SCAN HIGH OR EQUAL");
	return fdcError(val);
}
static uns8
fdcRecalibrate(uns8 val)
{
	switch (cntBytes++)
	{
		case 0:
			FdcStm |= STM_BUSY;
			DEB(Msg(M_DEBUG, "FDC: RECALIBRATE"););
			break;
		case 1:
			drvSelect((val & ST3_DRIVE) | (DRIVE->side << 2), 1);
			DEB(Msg(M_DEBUG, "      DS %d", DRIVE->id););
			st0 = DRIVE->id | (DRIVE->side << 2);
			st1 = 0;
			st2 = 0;
			if (DRIVE_READY)
			{
				drvSeek(0);
			}
			else
			{
				st0 |= ST0_STOPPED | ST0_NOTRDY;
			}
			st0 |= ST0_SEEKEND;
			intPending = 1;
			RESET_FDC_STATE;
			break;
	}
	return val;
}
static uns8
fdcSenseIntStatus(uns8 val)
{
	switch (cntBytes++)
	{
		case 0:
			DEB(Msg(M_DEBUG, "FDC: SENSE INTERRUPT STATUS"););
			FdcStm |= (STM_DIR | STM_BUSY);
			break;
		case 1:
			if (intPending)
			{
				intPending = 0;
				val = st0;
			}
			else
			{
				val = ST0_INVALCMD;
				RESET_FDC_STATE;
			}
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			break;
		case 2:
			val = DRIVE->track;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			RESET_FDC_STATE;
			break;
	}
	return val;
}
static uns8
fdcSpecify(uns8 val)
{
	switch (cntBytes++)
	{
		case 0:
			DEB(Msg(M_DEBUG, "FDC: SPECIFY"););
			FdcStm |= STM_BUSY;
			break;
		case 1:
			DEB(Msg(M_DEBUG, "      SRT %d, HUT %d", val >> 4, val & 0x0f););
			break;
		case 2:
			DEB(Msg(M_DEBUG, "      HLT %d, ND %d", val >> 1, val & 0x01););
			RESET_FDC_STATE;
			break;
	}
	return val;
}
static uns8
fdcSenseDriveStatus(uns8 val)
{
	switch (cntBytes++)
	{
		case 0:
			FdcStm |= STM_BUSY;
			DEB(Msg(M_DEBUG, "FDC: SENSE DRIVE STATUS"););
			break;
		case 1:
			drvSelect(val, 0);
			DEB(Msg(M_DEBUG, "      S %d, DS %d", DRIVE->side, DRIVE->id););
			FdcStm |= STM_DIR;
			break;
		case 2:
			st3 = DRIVE->id | (DRIVE->side << 2);
			if (DRIVE->sides == 2)
			{
				st3 |= ST3_DS;
			}
			if (!DRIVE->track)
			{
				st3 |= ST3_TRK0;
			}
			if (DRIVE_READY)
			{
				st3 |= ST3_RDY;
			}
			if (!DISK->fn || DISK->readonly)
			{
				st3 |= ST3_WRPROT;
			}
			if (DRIVE->id > 1)
			{
				st3 |= ST3_FAULT;
			}
			val = st3;
			DEB(Msg(M_DEBUG, "      -> %02x", val););
			RESET_FDC_STATE;
			break;
	}
	return val;
}
static uns8
fdcSeek(uns8 val)
{
	switch (cntBytes++)
	{
		case 0:
			DEB(Msg(M_DEBUG, "FDC: SEEK"););
			FdcStm |= STM_BUSY;
			break;
		case 1:
			drvSelect(val, 0);
			DEB(Msg(M_DEBUG, "      S %d, DS %d", DRIVE->side, DRIVE->id););
			break;
		case 2:
			DEB(Msg(M_DEBUG, "      DTK %d", val););
			st0 = DRIVE->id | (DRIVE->side << 2);
			st1 = 0;
			st2 = 0;
			if (DRIVE_READY)
			{
				if (val < DRIVE->tracks)
				{
					drvSeek((unsigned int)val);
				}
				else
				{
					st0 |= ST0_NOTFINISH | ST0_EQFAULT;
				}
			}
			else
			{
				st0 |= ST0_STOPPED | ST0_NOTRDY;
			}
			st0 |= ST0_SEEKEND;
			intPending = 1;
			RESET_FDC_STATE;
			break;
	}
	return val;
}
static uns8
fdcError(uns8 val)
{
	switch (cntBytes++)
	{
		case 0:
			FdcStm |= STM_DIR;
			st0 = ST0_INVALCMD;
			st1 = 0;
			st2 = 0;
			break;
		case 1:
			val = st0;
			RESET_FDC_STATE;
			break;
	}
	return val;
}
static void
drvSelect(uns8 new, int image)
{
	DRIVE = &drives[new & ST3_DRIVE];
	DRIVE_LED_ON;
	if (image && !DISK->fn)
	{
		(void)FdcInsertDisk((int)DRIVE->id);
	}
	if (DRIVE->side != (unsigned int)((new & ST3_SIDE) >> 2))
	{
		DRIVE->side = (new & ST3_SIDE) >> 2;
		if (DISK->fn)
		{
			drvSeek(DRIVE->track);
		}
	}
}
static void
drvSeek(unsigned int track)
{
	DRIVE->track = track;
	if ((DRIVE->side && DISK->sides == 1) || (DRIVE->track >= DISK->tracks))
	{
		DEB(Msg(M_DEBUG, "FDC: side/track out of range");)
		DRIVE->trackp = (trackInfo *)-1;
	}
	else
	{
		DRIVE->trackp = (trackInfo *)
							(DISK->mmp + sizeof(diskInfo) +
							 DRIVE->track * DISK->tracksize * DISK->sides +
							 DRIVE->side * DISK->tracksize);
		if (DRIVE->trackp->spt)
		{
			DRIVE->sectorp = (uns8 *)(DRIVE->trackp) +
							 sizeof(trackInfo) +
							 DRIVE->sector * (128U << DRIVE->trackp->bps);
		}
		else
		{
			DEB(Msg(M_DEBUG, "FDC: track not formatted");)
			DRIVE->sectorp = (uns8 *)-1;
		}
	}
}
static void
drvNextSector(void)
{
	assert(DRIVE->sectorp != (uns8 *)-1);
	if (++(DRIVE->sector) >= DRIVE->trackp->spt)
	{
		DRIVE->sector = 0;
		DRIVE->sectorp = (uns8 *)(DRIVE->trackp) + sizeof(trackInfo);
		st1 |= ST1_ENDCYL;
	}
	else
	{
		DRIVE->sectorp = (uns8 *)(DRIVE->trackp) + sizeof(trackInfo) +
						 (DRIVE->sector * (128U << DRIVE->trackp->bps));
		st1 &= ~ST1_ENDCYL;
	}
}
static sectorInfo *
drvFindSector(void)
{
	int i;
	assert(DISK->fn != (char *)0);
	if (DRIVE->trackp != (trackInfo *)-1 && DRIVE->sectorp != (uns8 *)-1)
	{
		for (i = 0; i < DRIVE->trackp->spt; i++)
		{
			if (!memcmp(&(DRIVE->trackp->si[DRIVE->sector]), &si, 4))
			{
				return &(DRIVE->trackp->si[DRIVE->sector]);
			}
			drvNextSector();
		}
	}
	return (sectorInfo *)0;
}
static int
diskOpenImage(int d, char *fname)
{
	driveDef *drive = &drives[d];
	diskDef *disk = (diskDef *)&(drive->disk);
	int readonly = 0;
	size_t size;
	diskInfo *diskinfo;
	assert(fname != (char *)0);
	diskCloseImage(d);
	if (drives[1 - d].disk.fn && !strcmp(fname, drives[1 - d].disk.fn))
	{
		Msg(M_WARN, "disk image <%s> allready assigned to drive %c",
			fname, 'B' - (char)d);
		return -1;
	}
	if ((disk->mmp = (uns8 *)Mmap(fname, 1, &readonly, &size)) == (uns8 *)-1)
	{
		return -1;
	}
	diskinfo = (diskInfo *)disk->mmp;
	if (memcmp(diskinfo->magic, "MV - CPC", 8))
	{
		if (memcmp(diskinfo->magic, "EXTENDED", 8))
		{
			Msg(M_WARN, "disk image is not in DSK format");
		}
		else
		{
			Msg(M_WARN, "XZX does not support extended DSK format");
		}
		Munmap(disk->mmp, size);
		return -1;
	}
	disk->fn = fname;
	disk->readonly = readonly;
	disk->mmsize = size;
	disk->tracks = diskinfo->tracks;
	disk->sides = diskinfo->sides;
	disk->tracksize = diskinfo->tracksize[0] | (diskinfo->tracksize[1] << 8);
	drive->side = 0;
	drive->track = 0;
	drive->sector = 0;
	drive->trackp = (trackInfo *)(disk->mmp + sizeof(diskInfo));
	drive->sectorp = (uns8 *)-1;
	if ((drive == DRIVE) && (FdcStm & STM_BUSY))
	{
		intPending = 0;
		RESET_FDC_STATE;
	}
	DEB(Msg(M_DEBUG,
			"FDC: %d tracks, %d sides, %d sectors, %d bytes, read%s",
			disk->tracks, disk->sides, drive->trackp->spt,
			128 << drive->trackp->bps, disk->readonly ? "only" : "write"););
	return 0;
}
static void
diskCloseImage(int d)
{
	driveDef *drive = &drives[d];
	diskDef *disk = (diskDef *)&(drive->disk);
	if (disk->fn)
	{
		assert(disk->mmp != (uns8 *)0);
		Munmap((void *)disk->mmp, disk->mmsize);
		free(disk->fn);
		disk->fn = (char *)0;
		if ((drive == DRIVE) && (FdcStm & STM_BUSY))
		{
			intPending = 0;
			RESET_FDC_STATE;
		}
	}
}
