/********************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation
 * for any purpose is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that the name
 * of the copyright holder not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.  The
 * copyright holder makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: Erik Kunze
 *******************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: machine.c,v 1.27 2000/11/26 18:57:02 erik Rel $";
#endif
#include <assert.h>
#include "cpu.h"
#include "resource.h"
#include "mem.h"
#include "util.h"
#ifdef XZX_DIDAKTIK
#include "didaktik.h"
#endif
#include "emul.h"
#include "fdc.h"
#ifdef XZX_IF1
#include "if1.h"
#endif
#ifdef XZX_KMOUSE
#include "kmouse.h"
#endif
#include "menu.h"
#ifdef XZX_MF
#include "multiface.h"
#endif
#ifdef XZX_MGT
#include "mgt.h"
#endif
#ifdef XZX_PENTAGON
#include "pentagon.h"
#endif
#include "plus3.h"
#include "sbar.h"
#ifdef XZX_SCORPION
#include "scorpion.h"
#endif
#include "spec48.h"
#include "spec128.h"
#ifdef XZX_TRDOS
#include "trdos.h"
#endif
#include "machine.h"
MachineSettings Machines[NMACHINES];
MachineSettings *Machine;
void
MachineInit(void)
{
	assert(CFGGET(machine) < NMACHINES);
	Machine = &Machines[CFGGET(machine)];
	Z80_Init();
	MemoryInit();
	Spec48Init();
	Spec128Init();
	Plus3Init();
#ifdef XZX_PENTAGON
	PentagonInit();
#endif
#ifdef XZX_SCORPION
	ScorpionInit();
#endif
#ifdef XZX_DIDAKTIK
	DidaktikInit();
#endif
	Machine->initPages();
	SbarSetModel(Machine->name);
	FdcInit();
#ifdef XZX_IF1
	If1Init();
#endif
#ifdef XZX_MF
	MultifaceInit();
#endif
#ifdef XZX_TRDOS
	TrdosInit();
#endif
#ifdef XZX_MGT
	MgtInit();
#endif
#ifdef XZX_KMOUSE
	KmouseInit();
#endif
	MenuFileDiskUpdate();
}
void
MachineSelect(int model, int reset)
{
	assert(model < NMACHINES);
	if (CFGGET(machine) != model)
	{
		CFGSET(machine, model);
		Machine = &Machines[model];
		SbarSetModel(Machine->name);
		EmulSetSpeed(CFGGET(speed));
#ifdef XZX_IF1
		if (CFGGET(if1) >= 0 && CFGGET(machine) > SP_128)
		{
			CFGSET(if1, 0);
			MenuFileCartridgeUpdate();
		}
#endif
#ifdef XZX_MF
		if (CFGGET(mf1) > 0 && CFGGET(machine) > SP_3)
		{
			CFGSET(mf1, 0);
		}
#endif
#ifdef XZX_TRDOS
		if (CFGGET(trdos) > 0 && CFGGET(machine) > SP_128)
		{
			CFGSET(trdos, 0);
		}
#endif
#ifdef XZX_MGT
		if (CFGGET(mgt) > 0 && CFGGET(machine) > SP_128)
		{
			CFGSET(mgt, 0);
		}
#endif
		MenuFileDiskUpdate();
#ifdef XZX_KMOUSE
		if (CFGGET(kmouse) > 0 && !PENTAGON_OR_SCORPION)
		{
			CFGSET(kmouse, 0);
			KmouseOff();
		}
#endif
	}
	if (reset)
	{
		Machine->reset();
	}
	else
	{
		Machine->initPages();
	}
}
