/********************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation
 * for any purpose is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that the name
 * of the copyright holder not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.  The
 * copyright holder makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 *******************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#if defined(sgi) && !defined(NAS_AUDIO) && !defined(OSS_AUDIO)
#ifndef lint
static char rcsid[] = "$Id: irix.c,v 1.15 2000/01/22 15:10:12 erik Rel $";
#endif
#include <sys/types.h>
#include <signal.h>
#include <dmedia/audio.h>
#include "debug.h"
#include "resource.h"
#include "util.h"
#include "main.h"
#include "audio.h"
#if HAVE_al_FUNCS
#define ALopenport		alOpenPort
#define ALcloseport		alClosePort
#define ALfreeconfig	alFreeConfig
#define ALgetfillable	alGetFillable
#define ALwritesamps	alWriteFrames
#endif
static ALport audioPort = (ALport)0;
static sigset_t blockingSet;
int
OsdOpenAudioDevice(char *device, int *sample, int *stereo)
{
	ALconfig audioConfig;
#if HAVE_al_FUNCS
	ALpv pvs[4];
	int alIsBusy = 0;
	pvs[0].param = AL_MAX_PORTS;
	pvs[1].param = AL_UNUSED_PORTS;
	if (alGetParams(AL_SYSTEM, pvs, 2) < 0)
	{
		return -1;
	}
	if (pvs[1].value.i < 1)
	{
		Msg(M_WARN, "audio device in use");
		return -1;
	}
	if (pvs[1].value.i < pvs[0].value.i)
	{
		pvs[0].param = AL_RATE;
		if (alGetParams(AL_DEFAULT_OUTPUT,pvs,1) < 0)
		{
			return -1;
		}
		*sample = pvs[0].value.i;
		alIsBusy = 1;
	}
	if ((audioConfig = alNewConfig()) == (ALconfig)0)
	{
		return -1;
	}
	if (alSetChannels(audioConfig, AL_STEREO) < 0
		|| alSetQueueSize(audioConfig, 10000L) < 0
		|| alSetSampFmt(audioConfig, AL_SAMPFMT_TWOSCOMP) < 0
		|| alSetWidth(audioConfig,AL_SAMPLE_8) < 0)
	{
		Msg(M_ERR, "couldn't set channel-specific parameters");
		goto error;
	}
	*stereo = 1;
	pvs[0].param = AL_MASTER_CLOCK;
	pvs[0].value.i = AL_CRYSTAL_MCLK_TYPE;
	pvs[1].param = AL_RATE;
	pvs[1].value.i = *sample;
	if (!alIsBusy)
	{
		if (alSetParams(AL_DEFAULT_OUTPUT, pvs, 2) < 0)
		{
			Msg(M_ERR, "couldn't set audio-device parameters");
			goto error;
		}
	}
#else
	long param[4];
	param[0] = AL_INPUT_COUNT;
	param[2] = AL_OUTPUT_COUNT;
	(void)ALgetparams(AL_DEFAULT_DEVICE, param, 4);
	if (param[1] || param[3])
	{
		Msg(M_WARN, "audio device in use");
		return -1;
	}
	if ((audioConfig = ALnewconfig()) == (ALconfig)0)
	{
		return -1;
	}
	if (ALsetchannels(audioConfig, AL_STEREO) < 0
		|| ALsetqueuesize(audioConfig, 10000L) < 0
		|| ALsetsampfmt(audioConfig, AL_SAMPFMT_TWOSCOMP) < 0
		|| ALsetwidth(audioConfig, AL_SAMPLE_8) < 0)
	{
		Msg(M_ERR, "couldn't set channel-specific parameters");
		goto error;
	}
	*stereo = 1;
	param[0] = AL_OUTPUT_RATE;
	param[1] = (long)*sample;
	param[2] = AL_INPUT_RATE;
	param[3] = (long)*sample;
	if (ALsetparams(AL_DEFAULT_DEVICE, param, 4) < 0)
	{
		Msg(M_ERR, "couldn't set audio-device parameters");
		goto error;
	}
#endif
	if ((audioPort = ALopenport(Version, "w", audioConfig)) == (ALport)0)
	{
		Msg(M_ERR, "couldn't open audio device for writing");
		goto error;
	}
	(void)ALfreeconfig(audioConfig);
	sigemptyset(&blockingSet);
	sigaddset(&blockingSet, SIGALRM);
	return 0;
  error:
	if (audioPort != (ALport)0)
	{
		(void)ALcloseport(audioPort);
		audioPort = (ALport)0;
	}
	(void)ALfreeconfig(audioConfig);
	return -1;
}
void
OsdCloseAudioDevice(void)
{
	if (audioPort != (ALport)0)
	{
		(void)ALcloseport(audioPort);
		audioPort = (ALport)0;
	}
}
int
OsdDrainAudioDevice(void)
{
	return 0;
}
size_t
OsdGetAudioFreespace(void)
{
	return ((size_t)ALgetfillable(audioPort));
}
size_t
OsdPlayAudioBuffer(unsigned char *buf, size_t len)
{
	sigset_t oldSet;
	size_t written, i;
	for (i = 0; i < len; i++)
	{
		buf[i] ^= 0x80;
	}
	(void)sigprocmask(SIG_BLOCK, &blockingSet, &oldSet);
	if ((written = ALwritesamps(audioPort, buf, (long)len)) < 0)
	{
		Msg(M_ERR, "couldn't write to audio device");
	}
	else
	{
		written = len;
	}
	if (!sigismember(&oldSet, SIGALRM))
	{
		(void)sigprocmask(SIG_UNBLOCK, &blockingSet, (sigset_t *)0);
	}
	return written;
}
#endif
