/********************************************************************************
 * Copyright (c) Des Herriott 1993, 1994
 *               Erik Kunze   1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation
 * for any purpose is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that the name
 * of the copyright holder not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.  The
 * copyright holder makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: Des Herriott
 *          Erik Kunze
 *******************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifdef REAL_AUDIO
#ifndef lint
static char speaker_c[] = "$Id: speaker.c,v 1.13 2000/11/18 16:33:17 erik Rel $";
#endif
#include <stdlib.h>
#include <assert.h>
#include "cpu.h"
#include "debug.h"
#include "resource.h"
#include "io.h"
#include "util.h"
#include "audio.h"
#ifdef DEBUG
#define DEB(x)			{ if (CFGGET(debug) & D_AUDIO) { x } }
#else
#define DEB(x)
#endif
static void speakerUpdate(unsigned int);
void (*SpeakerOutByte)(uns8);
static unsigned int lastPos, lastRem, accum;
static unsigned char lastVal;
static unsigned char *buf;
void
SpeakerInit(void)
{
	lastPos = 0;
	lastRem = 0;
	accum = 0;
	if (buf)
	{
		free(buf);
	}
	buf = Malloc((size_t)(2 * Audio.samplesPerFrame), "SpeakerInit");
	Audio.spBuf = buf;
}
void
SpeakerOut(uns8 newval)
{
	unsigned int pos, rem;
	unsigned char val = ((newval & (B_SPEAKER | B_MIC)) << 3) - 96;
	if (val != lastVal)
	{
		rem = TSTATES % Audio.tstatesPerSample;
		if ((pos = TSTATES / Audio.tstatesPerSample) != lastPos)
		{
			if (pos > Audio.samplesPerFrame)
			{
				pos = Audio.samplesPerFrame;
			}
			speakerUpdate(pos);
			accum = lastVal * rem;
		}
		else
		{
			accum += lastVal * (rem - lastRem);
		}
		lastRem = rem;
		lastVal = val;
	}
}
void
SpeakerFillBuffer(void)
{
	if (lastPos < Audio.samplesPerFrame)
	{
		speakerUpdate(Audio.samplesPerFrame);
	}
}
void
SpeakerResetBuffer(void)
{
	lastPos = 0;
	lastRem = 0;
	accum = 0;
}
static void
speakerUpdate(unsigned int pos)
{
	assert(pos <= Audio.samplesPerFrame);
	buf[lastPos++] = (accum + lastVal * (Audio.tstatesPerSample - lastRem)) /
					 Audio.tstatesPerSample;
	for (; lastPos < pos; lastPos++)
	{
		buf[lastPos] = lastVal;
	}
}
#endif
