/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: menu.c,v 4.88 2000/11/04 21:00:53 erik Rel $";
#endif
#include <Xm/Xm.h>
#include <Xm/CascadeBG.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/SeparatoG.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "cpu.h"
#include "resource.h"
#include "util.h"
#ifdef XZX_DIDAKTIK
#include "didaktik.h"
#endif
#include "fdc.h"
#ifdef XZX_IF1
#include "if1.h"
#endif
#ifdef HAVE_XPM
#include "kbd.h"
#endif
#ifdef XZX_MGT
#include "mgt.h"
#endif
#include "monitor.h"
#include "options.h"
#include "poke.h"
#include "sbar.h"
#include "scr.h"
#include "snapshot.h"
#ifdef XZX_TRDOS
#include "trdos.h"
#endif
#include "ui.h"
#include "vtp.h"
#include "xwin.h"
#include "menu.h"
typedef struct _menuItem {
	const char *label;
	WidgetClass *class;
	char mnemonic;
	const char *accelerator;
	const char *accelText;
	void (*callback)(Widget, XtPointer, XtPointer);
	XPointer callbackData;
	struct _menuItem *subItems;
} menuItem;
#define MENUBEGIN(a)					static menuItem a[] = {
#define MENUITEM(a,b,c,d,e,f,g,h) 		{ (const char *)a,(WidgetClass *)b,(char)c, \
										  (const char *)d,(const char *)e,(XtCallbackProc)f, \
										  (XPointer)g,(struct _menuItem *)h },
#define MENUSEP							MENUITEM("sep",&xmSeparatorGadgetClass,0,0,0,0,0,0)
#define MENUEND							MENUITEM(0,0,0,0,0,0,0,0) };
#define PB_CLASS						&xmPushButtonGadgetClass
#define TB_CLASS						&xmToggleButtonGadgetClass
#define CB_CLASS						&xmCascadeButtonGadgetClass
#define XtNull 							(XtPointer)0
#define XtAddCallback(w,cb,f,val)		XtAddCallback(w,cb,f,(XtPointer)val)
enum { HELP_COPYRIGHT, HELP_README, HELP_FAQ, HELP_KEYPIC };
static Widget menuCreateMenu(Widget, int, const char *, char, Boolean, menuItem *);
static void menuFileSnapshotRead(void);
static void menuFileSnapshotWrite(void);
static void menuFileDiskInsert(Widget, XtPointer, XtPointer);
static void menuFileDiskRemove(Widget, XtPointer, XtPointer);
#ifdef XZX_IF1
static void menuFileCartridgeInsert(Widget, XtPointer, XtPointer);
static void menuFileCartridgeRemove(Widget, XtPointer, XtPointer);
#endif
static void menuFileScreendumpRead(void);
static void menuFileScreendumpWrite(void);
static void menuFileQuit(void);
static void menuMiscReset(void);
static void menuMiscNmi(void);
static void menuMiscPause(Widget, XtPointer, XtPointer);
static void menuMiscMonitor(Widget, XtPointer, XtPointer);
static void menuMiscVtp(Widget, XtPointer, XtPointer);
static void menuOptionsSaveOnExit(Widget, XtPointer, XtPointer);
static void menuOptionsSaveNow(void);
static void menuHelp(Widget, XtPointer, XtPointer);
static void menuHelpCallback(Widget, XtPointer, XtPointer);
MENUBEGIN(diskInsertItems)
	MENUITEM("A", PB_CLASS, 'A', "Ctrl<Key>a", "Ctrl+A", menuFileDiskInsert, 0, 0)
	MENUITEM("B", PB_CLASS, 'B', "Ctrl<Key>b", "Ctrl+B", menuFileDiskInsert, 1, 0)
#if DISK_DRIVES > 2
	MENUITEM("C", PB_CLASS, 'C', "Ctrl<Key>c", "Ctrl+C", menuFileDiskInsert, 2, 0)
	MENUITEM("D", PB_CLASS, 'D', "Ctrl<Key>d", "Ctrl+D", menuFileDiskInsert, 3, 0)
#endif
MENUEND
MENUBEGIN(diskRemoveItems)
	MENUITEM("A", PB_CLASS, 'A', 0, 0, menuFileDiskRemove, 0, 0)
	MENUITEM("B", PB_CLASS, 'B', 0, 0, menuFileDiskRemove, 1, 0)
#if DISK_DRIVES > 2
	MENUITEM("C", PB_CLASS, 'C', 0, 0, menuFileDiskRemove, 2, 0)
	MENUITEM("D", PB_CLASS, 'D', 0, 0, menuFileDiskRemove, 3, 0)
#endif
MENUEND
#ifdef XZX_IF1
MENUBEGIN(mdrInsertItems)
	MENUITEM("1", PB_CLASS, '1', "Ctrl<Key>1", "Ctrl+1", menuFileCartridgeInsert, 0, 0)
#if IF1_DRIVES > 1
	MENUITEM("2", PB_CLASS, '2', "Ctrl<Key>2", "Ctrl+2", menuFileCartridgeInsert, 1, 0)
#endif
#if IF1_DRIVES > 2
	MENUITEM("3", PB_CLASS, '3', "Ctrl<Key>3", "Ctrl+3", menuFileCartridgeInsert, 2, 0)
#endif
#if IF1_DRIVES > 3
	MENUITEM("4", PB_CLASS, '4', "Ctrl<Key>4", "Ctrl+4", menuFileCartridgeInsert, 3, 0)
#endif
#if IF1_DRIVES > 4
	MENUITEM("5", PB_CLASS, '5', "Ctrl<Key>5", "Ctrl+5", menuFileCartridgeInsert, 4, 0)
#endif
#if IF1_DRIVES > 5
	MENUITEM("6", PB_CLASS, '6', "Ctrl<Key>6", "Ctrl+6", menuFileCartridgeInsert, 5, 0)
#endif
#if IF1_DRIVES > 6
	MENUITEM("7", PB_CLASS, '7', "Ctrl<Key>7", "Ctrl+7", menuFileCartridgeInsert, 6, 0)
#endif
#if IF1_DRIVES > 7
	MENUITEM("8", PB_CLASS, '8', "Ctrl<Key>8", "Ctrl+8", menuFileCartridgeInsert, 7, 0)
#endif
MENUEND
MENUBEGIN(mdrRemoveItems)
	MENUITEM("1", PB_CLASS, '1', 0, 0, menuFileCartridgeRemove, 0, 0)
#if IF1_DRIVES > 1
	MENUITEM("2", PB_CLASS, '2', 0, 0, menuFileCartridgeRemove, 1, 0)
#endif
#if IF1_DRIVES > 2
	MENUITEM("3", PB_CLASS, '3', 0, 0, menuFileCartridgeRemove, 2, 0)
#endif
#if IF1_DRIVES > 3
	MENUITEM("4", PB_CLASS, '4', 0, 0, menuFileCartridgeRemove, 3, 0)
#endif
#if IF1_DRIVES > 4
	MENUITEM("5", PB_CLASS, '5', 0, 0, menuFileCartridgeRemove, 4, 0)
#endif
#if IF1_DRIVES > 5
	MENUITEM("6", PB_CLASS, '6', 0, 0, menuFileCartridgeRemove, 5, 0)
#endif
#if IF1_DRIVES > 6
	MENUITEM("7", PB_CLASS, '7', 0, 0, menuFileCartridgeRemove, 6, 0)
#endif
#if IF1_DRIVES > 7
	MENUITEM("8", PB_CLASS, '8', 0, 0, menuFileCartridgeRemove, 7, 0)
#endif
MENUEND
#endif
MENUBEGIN(fileItems)
	MENUITEM("Load snapshot...", PB_CLASS, 'L', "<Key>F3",    "F3",     menuFileSnapshotRead,    0, 0              )
	MENUITEM("Save snapshot...", PB_CLASS, 'S', "<Key>F2",    "F2",     menuFileSnapshotWrite,   0, 0              )
	MENUSEP
	MENUITEM("Insert disk",      CB_CLASS, 'I', 0,            0,        0,                       0, diskInsertItems)
	MENUITEM("Remove disk",      CB_CLASS, 'R', 0,            0,        0,                       0, diskRemoveItems)
	MENUSEP
#ifdef XZX_IF1
	MENUITEM("Insert cartridge", CB_CLASS, 'n', 0,            0,        0,                       0, mdrInsertItems )
	MENUITEM("Remove cartridge", CB_CLASS, 'e', 0,            0,        0,                       0, mdrRemoveItems )
	MENUSEP
#endif
	MENUITEM("Load screen...",   PB_CLASS, 'o', 0,            0,        menuFileScreendumpRead,  0, 0              )
	MENUITEM("Save screen...",   PB_CLASS, 'a', "<Key>F4",    "F4",     menuFileScreendumpWrite, 0, 0              )
	MENUSEP
	MENUITEM("Quit",             PB_CLASS, 'Q', "Ctrl<Key>q", "Ctrl+Q", menuFileQuit,            0, 0              )
MENUEND
MENUBEGIN(miscItems)
	MENUITEM("Reset",            PB_CLASS, 'R', "<Key>F5",    "F5",     menuMiscReset,           0, 0              )
	MENUITEM("NMI",              PB_CLASS, 'N', "<Key>F6",    "F6",     menuMiscNmi,             0, 0              )
	MENUITEM("Pause",            TB_CLASS, 'P', "Ctrl<Key>h", "Ctrl+H", menuMiscPause,           0, 0              )
	MENUSEP
	MENUITEM("Monitor",          TB_CLASS, 'M', "<Key>F7",    "F7",     menuMiscMonitor,         0, 0              )
	MENUITEM("Tape manager",     TB_CLASS, 'T', "<Key>F8",    "F8",     menuMiscVtp,             0, 0              )
	MENUITEM("Poke manager...",  PB_CLASS, 'o', 0,            0,        PokeManager,             0, 0              )
MENUEND
MENUBEGIN(optionItems)
	MENUITEM("Change...",        PB_CLASS, 'C', "<Key>F9",    "F9",     OptionsDialog,           0, 0              )
	MENUITEM("Save on exit",     TB_CLASS, 'x', 0,            0,        menuOptionsSaveOnExit,   0, 0              )
	MENUITEM("Save now",         PB_CLASS, 'n', 0,            0,        menuOptionsSaveNow,      0, 0              )
MENUEND
MENUBEGIN(helpItems)
#ifdef HAVE_XPM
	MENUITEM("Keyboard",         TB_CLASS, 'K', "<Key>F1",    "F1",     menuHelp,                HELP_KEYPIC,    0 )
#endif
	MENUITEM("Copyright",        PB_CLASS, 'C', 0,            0,        menuHelp,                HELP_COPYRIGHT, 0 )
	MENUITEM("Readme",           PB_CLASS, 'R', 0,            0,        menuHelp,                HELP_README,    0 )
	MENUITEM("FAQ",              PB_CLASS, 'F', 0,            0,        menuHelp,                HELP_FAQ,       0 )
	MENUSEP
	MENUITEM("About",            PB_CLASS, 'A', 0,            0,        UiShowAbout,             0,              0 )
MENUEND
static struct {
#ifdef XZX_IF1
	Widget mi[IF1_DRIVES];
	Widget mr[IF1_DRIVES];
#endif
	Widget di[DISK_DRIVES];
	Widget dr[DISK_DRIVES];
} menuFileItems;
static Boolean menuSaveOnExit = True;
Widget
MenuCreateBar(Widget parent)
{
	Widget bar, helpmenu;
	char name[80];
	int i;
	bar = XmCreateMenuBar(parent, (char *)"bar", (ArgList)0, 0);
	(void)menuCreateMenu(bar, XmMENU_PULLDOWN, "File", 0, False, fileItems);
	(void)menuCreateMenu(bar, XmMENU_PULLDOWN, "Misc", 0, False, miscItems);
	(void)menuCreateMenu(bar, XmMENU_PULLDOWN, "Options", 0, False, optionItems);
	helpmenu = menuCreateMenu(bar, XmMENU_PULLDOWN, "Help", 0, False, helpItems);
	XtVaSetValues(bar, XmNmenuHelpWidget, helpmenu, XtNull);
#ifdef XZX_IF1
	for (i = IF1_DRIVES; i; i--)
	{
		(void)sprintf(name, "*%s.%d", "Insert cartridge", i);
		menuFileItems.mi[i - 1] = XtNameToWidget(parent, name);
		(void)sprintf(name, "*%s.%d", "Remove cartridge", i);
		menuFileItems.mr[i - 1] = XtNameToWidget(parent, name);
	}
#endif
	for (i = DISK_DRIVES; i; i--)
	{
		(void)sprintf(name, "*%s.%c", "Insert disk", '@' + (char)i);
		menuFileItems.di[i - 1] = XtNameToWidget(parent, name);
		(void)sprintf(name, "*%s.%c", "Remove disk", '@' + (char)i);
		menuFileItems.dr[i - 1] = XtNameToWidget(parent, name);
	}
	(void)sprintf(name, "*%s", "Save on exit");
	XmToggleButtonGadgetSetState(XtNameToWidget(parent, name), menuSaveOnExit, False);
	return bar;
}
static Widget
menuCreateMenu(Widget parent, int menuType, const char *menuTitle, char menuMnemonic, Boolean tearOff, menuItem *items)
{
	Widget menu, cascade = (Widget)0, widget;
	XmString str;
	Arg args[2];
	Cardinal i;
	if (menuType == XmMENU_PULLDOWN || menuType == XmMENU_OPTION)
	{
		menu = XmCreatePulldownMenu(parent, (char *)menuTitle, (ArgList)0, 0);
	}
	else
	{
		assert(menuType == XmMENU_POPUP);
		menu = XmCreatePopupMenu(parent, (char *)menuTitle, (ArgList)0, 0);
	}
	if (tearOff == True)
	{
		XtVaSetValues(menu, XmNtearOffModel, XmTEAR_OFF_ENABLED, XtNull);
	}
	if (menuType == XmMENU_PULLDOWN)
	{
		i = 0;
		XtSetArg(args[i], XmNsubMenuId, menu); i++;
		if (menuMnemonic != '\0')
		{
			XtSetArg(args[i], XmNmnemonic, menuMnemonic); i++;
		}
		cascade = XmCreateCascadeButtonGadget(parent, (char *)menuTitle, args, i);
		XtManageChild(cascade);
	}
	else if (menuType == XmMENU_OPTION)
	{
		i = 0;
		XtSetArg(args[i], XmNsubMenuId, menu); i++;
		if (menuMnemonic != '\0')
		{
			XtSetArg(args[i], XmNmnemonic, menuMnemonic); i++;
		}
		cascade = XmCreateOptionMenu(parent, (char *)menuTitle, args, i);
		XtManageChild(cascade);
	}
	for (i = 0; items[i].label != (const char *)0; i++)
	{
		if (items[i].subItems != (menuItem *)0)
		{
			assert(menuType != XmMENU_OPTION);
			widget = menuCreateMenu(menu, XmMENU_PULLDOWN, items[i].label, items[i].mnemonic, tearOff, items[i].subItems);
		}
		else
		{
			widget = XtVaCreateManagedWidget(items[i].label, *items[i].class, menu, XtNull);
		}
		if (items[i].mnemonic != '\0')
		{
			XtVaSetValues(widget, XmNmnemonic, items[i].mnemonic, XtNull);
		}
		if (items[i].accelerator != (const char *)0)
		{
			str = XmStringCreateLtoR((char *)(items[i].accelText), XmFONTLIST_DEFAULT_TAG);
			XtVaSetValues(widget, XmNaccelerator, items[i].accelerator, XmNacceleratorText, str, XtNull);
			XmStringFree(str);
		}
		if (items[i].callback != (XtPointer)0)
		{
			XtAddCallback(widget,
						  (items[i].class == &xmToggleButtonWidgetClass ||
						   items[i].class == &xmToggleButtonGadgetClass) ?
						  XmNvalueChangedCallback :
						  XmNactivateCallback ,
						  items[i].callback, items[i].callbackData);
		}
	}
	return (menuType == XmMENU_POPUP ? menu : cascade);
}
static void
menuFileSnapshotRead(void)
{
	int types[4] = { FT_Z80, FT_SLT, FT_SNA, FT_END };
	char *fname;
	if ((fname = UiFileSelector("Load Snapshot", 1, types)) != (char *)0)
	{
		(void)SnapshotRead(fname);
		free(fname);
	}
}
static void
menuFileSnapshotWrite(void)
{
	int types[4] = { FT_Z80, FT_SLT, FT_SNA, FT_END };
	char *fname;
	if (CFGGET(machine) > SP_48_3)
	{
		types[2] = FT_END;
	}
	if ((fname = UiFileSelector("Save Snapshot", 0, types)) != (char *)0)
	{
		(void)SnapshotWrite(fname);
		free(fname);
	}
}
static void
menuFileDiskInsert(Widget w, XtPointer client, XtPointer call)
{
	if (CFGGET(machine) == SP_3)
	{
		(void)FdcInsertDisk((int)client);
	}
#ifdef XZX_DIDAKTIK
	else if (CFGGET(machine) == SP_DIDAKTIK)
	{
		(void)DidaktikInsertDisk((int)client);
	}
#endif
#ifdef XZX_TRDOS
	else if (CFGGET(trdos) > 0 || PENTAGON_OR_SCORPION)
	{
		(void)TrdosInsertDisk((int)client);
	}
#endif
#ifdef XZX_MGT
	else if (CFGGET(mgt) > 0)
	{
		(void)MgtInsertDisk((int)client);
	}
#endif
}
static void
menuFileDiskRemove(Widget w, XtPointer client, XtPointer call)
{
	if (CFGGET(machine) == SP_3)
	{
		FdcEjectDisk((int)client);
	}
#ifdef XZX_DIDAKTIK
	else if (CFGGET(machine) == SP_DIDAKTIK)
	{
		DidaktikEjectDisk((int)client);
	}
#endif
#ifdef XZX_TRDOS
	else if (CFGGET(trdos) > 0 || PENTAGON_OR_SCORPION)
	{
		TrdosEjectDisk((int)client);
	}
#endif
#ifdef XZX_MGT
	else if (CFGGET(mgt) > 0)
	{
		MgtEjectDisk((int)client);
	}
#endif
}
void
MenuFileDiskUpdate(void)
{
	char *(*func)(int) = 0;
	Boolean ins, rem;
	int drives = 0, i;
	if (CFGGET(machine) == SP_3)
	{
		func = FdcDiskName;
		drives = 2;
	}
#ifdef XZX_DIDAKTIK
	else if (CFGGET(machine) == SP_DIDAKTIK)
	{
		func = DidaktikDiskName;
		drives = 2;
	}
#endif
#ifdef XZX_TRDOS
	else if (CFGGET(trdos) > 0 || PENTAGON_OR_SCORPION)
	{
		func = TrdosDiskName;
		drives = 4;
	}
#endif
#ifdef XZX_MGT
	else if (CFGGET(mgt) > 0)
	{
		func = MgtDiskName;
		drives = 2;
	}
#endif
	for (i = 0; i < DISK_DRIVES; i++)
	{
		if (i >= drives)
		{
			ins = False;
			rem = False;
		}
		else
		{
			assert(func != 0);
			ins = True;
			rem = func(i) != (char *)0 ? True : False;
		}
		XtSetSensitive(menuFileItems.di[i], ins);
		XtSetSensitive(menuFileItems.dr[i], rem);
	}
	SbarSetLed(drives > 0 ? LED_SHOW : LED_HIDE);
}
#ifdef XZX_IF1
static void
menuFileCartridgeInsert(Widget w, XtPointer client, XtPointer call)
{
	assert(CFGGET(if1) > 0);
	(void)If1InsertCart((int)client);
}
#endif
#ifdef XZX_IF1
static void
menuFileCartridgeRemove(Widget w, XtPointer client, XtPointer call)
{
	assert(CFGGET(if1) > 0);
	If1EjectCart((int)client);
}
#endif
#ifdef XZX_IF1
void
MenuFileCartridgeUpdate(void)
{
	Boolean ins, rem;
	int i;
	for (i = 0; i < IF1_DRIVES; i++)
	{
		if (CFGGET(if1) <= 0)
		{
			ins = False;
			rem = False;
		}
		else
		{
			ins = True;
			rem = If1CartName(i) != (char *)0 ? True : False;
		}
		XtSetSensitive(menuFileItems.mi[i], ins);
		XtSetSensitive(menuFileItems.mr[i], rem);
	}
}
#endif
static void
menuFileScreendumpRead(void)
{
	int types[2] = { FT_SCR, FT_END };
	char *fname;
	if ((fname = UiFileSelector("Load Screenshot", 1, types)) != (char *)0)
	{
		(void)ScrRead(fname);
		free(fname);
	}
}
static void
menuFileScreendumpWrite(void)
{
	int types[2] = { FT_SCR, FT_END };
	char *fname;
	if ((fname = UiFileSelector("Save Screenshot", 0, types)) != (char *)0)
	{
		(void)ScrWrite(fname);
		free(fname);
	}
}
static void
menuFileQuit(void)
{
	if (UiShowMessage(DIALOG_ASK, "Do you really want to quit?", (const char *)0, (const char *)0, DIALOG_CANCEL) == DIALOG_OK)
	{
		if (menuSaveOnExit == True
			&& UiShowMessage(DIALOG_ASK, "Save the current settings?", (const char *)0, (const char *)0,
							 DIALOG_CANCEL) == DIALOG_OK)
		{
			if (ResourceWrite() < 0)
			{
				(void)UiShowMessage(DIALOG_ERROR, "Could not save sattings!", (const char *)0, (const char *)0, DIALOG_OK);
			}
		}
		Quit(0);
	}
}
static void
menuMiscReset(void)
{
	Machine->reset();
}
static void
menuMiscNmi(void)
{
	Machine->nmi();
}
static void
menuMiscPause(Widget w, XtPointer client, XtPointer call)
{
	static int paused;
	if (((XmToggleButtonCallbackStruct *)call)->set == True)
	{
		char *title, *saveTitle;
		XwinData *xdp = &XwinDatas;
		(void)UiEnter();
		paused = 1;
		XtVaGetValues(xdp->topLevel, XtNtitle, &title, XtNull);
		saveTitle = Strdup(title, "menuMiscPause");
		title = Malloc(strlen(saveTitle) + 10, "menuMiscPause");
		(void)sprintf(title, "%s (paused)", saveTitle);
		XtVaSetValues(xdp->topLevel, XtNtitle, title, XtNull);
		while (paused != 0)
		{
			XtAppProcessEvent(XtWidgetToApplicationContext(w), XtIMAll);
		}
		XtVaSetValues(xdp->topLevel, XtNtitle, saveTitle, XtNull);
		free(saveTitle);
		free(title);
		(void)UiLeave();
	}
	else
	{
		paused = 0;
	}
}
static void
menuMiscMonitor(Widget w, XtPointer client, XtPointer call)
{
	if (((XmToggleButtonCallbackStruct *)call)->set == True)
	{
		MonitorShow();
	}
	else
	{
		MonitorHide();
	}
}
static void
menuMiscVtp(Widget w, XtPointer client, XtPointer call)
{
	if (((XmToggleButtonCallbackStruct *)call)->set == True)
	{
		VtpShow();
	}
	else
	{
		VtpHide();
	}
}
static void
menuOptionsSaveOnExit(Widget w, XtPointer client, XtPointer call)
{
	menuSaveOnExit = ((XmToggleButtonCallbackStruct *)call)->set;
}
static void
menuOptionsSaveNow(void)
{
	int type;
	const char *msg;
	if (ResourceWrite() < 0)
	{
		type = DIALOG_ERROR;
		msg = "Could not save sattings!";
	}
	else
	{
		type = DIALOG_INFO;
		msg = "Settings saved successfully.";
	}
	(void)UiShowMessage(type, msg, (const char *)0, (const char *)0, DIALOG_OK);
}
static void
menuHelp(Widget w, XtPointer client, XtPointer call)
{
	const char *file;
	switch((int)client)
	{
		default:
		case HELP_COPYRIGHT:
			file = "COPYRIGHT";
			break;
		case HELP_README:
			file = "README";
			break;
		case HELP_FAQ:
			file = "FAQ";
			break;
#ifdef HAVE_XPM
		case HELP_KEYPIC:
			if (((XmToggleButtonCallbackStruct *)call)->set == True)
			{
				KbdShow(w);
			}
			else
			{
				KbdHide(w);
			}
			return;
#endif
	}
	if (UiShowTextFile(file, menuHelpCallback, (XtPointer)w) != 0)
	{
		(void)UiShowMessage(DIALOG_ERROR, "Could not open file!", (const char *)0, (const char *)0, DIALOG_OK);
	}
	else
	{
		XtSetSensitive(w, False);
	}
}
static void
menuHelpCallback(Widget w, XtPointer client, XtPointer call)
{
	XtSetSensitive((Widget)client, True);
}
