/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: options.c,v 1.15 2000/11/26 18:50:31 erik Rel $";
#endif
#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/LabelG.h>
#include <Xm/Notebook.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/Scale.h>
#include <Xm/SelectioB.h>
#include <Xm/SeparatoG.h>
#include <Xm/SSpinB.h>
#include <Xm/TextF.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>
#include <assert.h>
#include "debug.h"
#include "resource.h"
#include "machine.h"
#include "util.h"
#include "ui.h"
#ifdef AUDIO
#include "audio.h"
#endif
#include "emul.h"
#include "menu.h"
#include "tape.h"
#include "toolhelp.h"
#include "xwin.h"
#include "options.h"
enum { OPTIONS_NOP, OPTIONS_APPLY, OPTIONS_DISCARD };
typedef struct _optionsPage {
	int pageNumber;
	const char *pageLabel;
	int (*pageCreate)(Widget, struct _optionsPage *);
	void (*pageCallback)(Widget, XtPointer, XtPointer);
	void (*pageUpdate)(struct _optionsPage *);
	void (*pageCommit)(struct _optionsPage *);
	void *pageData;
} optionsPage;
#define XtNull 							(XtPointer)0
#define XtAddCallback(w,cb,f,val)		XtAddCallback(w,cb,f,(XtPointer)(val))
static void optionsDialogCallback(Widget, XtPointer, XtPointer);
static int optionsMiscCreate(Widget, optionsPage *);
static void optionsMiscCallback(Widget, XtPointer, XtPointer);
static void optionsMiscUpdate(optionsPage *);
static void optionsMiscCommit(optionsPage *);
static int optionsHardwareCreate(Widget, optionsPage *);
static void optionsHardwareCallback(Widget, XtPointer, XtPointer);
static void optionsHardwareUpdate(optionsPage *);
static void optionsHardwareCommit(optionsPage *);
static int optionsTapeCreate(Widget, optionsPage *);
static void optionsTapeCallback(Widget, XtPointer, XtPointer);
static void optionsTapeUpdate(optionsPage *);
static void optionsTapeCommit(optionsPage *);
#ifdef AUDIO
static int optionsSoundCreate(Widget, optionsPage *);
static void optionsSoundCallback(Widget, XtPointer, XtPointer);
static void optionsSoundUpdate(optionsPage *);
static void optionsSoundCommit(optionsPage *);
#endif
static int optionsJoystickCreate(Widget, optionsPage *);
static void optionsJoystickCallback(Widget, XtPointer, XtPointer);
static void optionsJoystickUpdate(optionsPage *);
static void optionsJoystickCommit(optionsPage *);
#ifdef XZX_IF1
static int optionsIf1Create(Widget, optionsPage *);
static void optionsIf1Callback(Widget, XtPointer, XtPointer);
static void optionsIf1Update(optionsPage *);
static void optionsIf1Commit(optionsPage *);
#endif
#ifdef PRINTER
static int optionsPrinterCreate(Widget, optionsPage *);
static void optionsPrinterCallback(Widget, XtPointer, XtPointer);
static void optionsPrinterUpdate(optionsPage *);
static void optionsPrinterCommit(optionsPage *);
#endif
#ifdef DEBUG
static int optionsDebugCreate(Widget, optionsPage *);
static void optionsDebugCallback(Widget, XtPointer, XtPointer);
static void optionsDebugUpdate(optionsPage *);
static void optionsDebugCommit(optionsPage *);
#endif
static optionsPage optionsPages[] = {
	{  1, "Misc",        optionsMiscCreate,     optionsMiscCallback,     optionsMiscUpdate,     optionsMiscCommit,     0 },
	{  2, "Hardware",    optionsHardwareCreate, optionsHardwareCallback, optionsHardwareUpdate, optionsHardwareCommit, 0 },
	{  3, "Tape",        optionsTapeCreate,     optionsTapeCallback,     optionsTapeUpdate,     optionsTapeCommit,     0 },
#ifdef AUDIO
	{  4, "Sound",       optionsSoundCreate,    optionsSoundCallback,    optionsSoundUpdate,    optionsSoundCommit,    0 },
#endif
	{  5, "Joystick",    optionsJoystickCreate, optionsJoystickCallback, optionsJoystickUpdate, optionsJoystickCommit, 0 },
#ifdef XZX_IF1
	{  6, "Interface I", optionsIf1Create,      optionsIf1Callback,      optionsIf1Update,      optionsIf1Commit,      0 },
#endif
#ifdef xPRINTER
	{  7, "Printer",     optionsPrinterCreate,  optionsPrinterCallback,  optionsPrinterUpdate,  optionsPrinterCommit,  0 },
#endif
#ifdef DEBUG
	{  8, "Debug",       optionsDebugCreate,    optionsDebugCallback,    optionsDebugUpdate,    optionsDebugCommit,    0 },
#endif
};
void
OptionsDialog(void)
{
	static Widget dialog = (Widget)0;
	static int answer;
	optionsPage *opp;
	if (dialog == (Widget)0)
	{
		XmString title, cancelLabel;
		Widget notebook, form, tab;
		Arg arg[7];
		Cardinal i;
		int j;
		title = XmStringCreateLtoR((char *)"XZX Options", XmFONTLIST_DEFAULT_TAG);
		cancelLabel = XmStringCreateLtoR((char *)"Discard", XmFONTLIST_DEFAULT_TAG);
		i = 0;
		XtSetArg(arg[i], XmNnoResize,            True                           ); i++;
		XtSetArg(arg[i], XmNautoUnmanage,        False                          ); i++;
		XtSetArg(arg[i], XmNdeleteResponse,      XmDO_NOTHING                   ); i++;
		XtSetArg(arg[i], XmNdialogStyle,         XmDIALOG_FULL_APPLICATION_MODAL); i++;
		XtSetArg(arg[i], XmNdialogTitle,         title                          ); i++;
		XtSetArg(arg[i], XmNcancelLabelString,   cancelLabel                    ); i++;
		XtSetArg(arg[i], XtNkeyboardFocusPolicy, XmPOINTER                      ); i++;
		assert(i <= XtNumber(arg));
		dialog = XmCreatePromptDialog(XwinDatas.topLevel, (char *)"options", arg, i);
		XmStringFree(title);
		XmStringFree(cancelLabel);
		XtUnmanageChild(XmSelectionBoxGetChild(dialog, XmDIALOG_COMMAND_TEXT));
		XtUnmanageChild(XmSelectionBoxGetChild(dialog, XmDIALOG_PROMPT_LABEL));
		XtUnmanageChild(XmSelectionBoxGetChild(dialog, XmDIALOG_OK_BUTTON));
		XtManageChild(XmSelectionBoxGetChild(dialog, XmDIALOG_APPLY_BUTTON));
		XtUnmanageChild(XmSelectionBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));
		XtAddCallback(dialog, XmNapplyCallback, optionsDialogCallback, &answer);
		XtAddCallback(dialog, XmNcancelCallback, optionsDialogCallback, &answer);
		i = 0;
		XtSetArg(arg[i], XmNuserData, &optionsPages); i++;
		assert(i <= XtNumber(arg));
		notebook = XmCreateNotebook(dialog, (char *)"notebook", arg, i);
		for (opp = optionsPages, j = 1; opp < &optionsPages[NELEM(optionsPages)]; opp++, j++)
		{
			opp->pageNumber = j;
			i = 0;
			XtSetArg(arg[i], XmNnotebookChildType, XmPAGE); i++;
			XtSetArg(arg[i], XmNpageNumber,        j     ); i++;
			XtSetArg(arg[i], XmNmarginHeight,      5     ); i++;
			XtSetArg(arg[i], XmNmarginWidth,       5     ); i++;
			assert(i <= XtNumber(arg));
			form = XmCreateForm(notebook, (char *)"form", arg, i);
			i = 0;
			XtSetArg(arg[i], XmNnotebookChildType, XmMAJOR_TAB); i++;
			XtSetArg(arg[i], XmNpageNumber,        j          ); i++;
			assert(i <= XtNumber(arg));
			tab = XmCreatePushButtonGadget(notebook, (char *)opp->pageLabel, arg, i);
			opp->pageCreate(form, opp);
			XtManageChild(tab);
			XtManageChild(form);
		}
		XtManageChild(notebook);
	}
	for (opp = optionsPages; opp < &optionsPages[NELEM(optionsPages)]; opp++)
	{
		opp->pageUpdate(opp);
	}
	UiShowDialog(dialog, &answer);
}
static void
optionsDialogCallback(Widget w, XtPointer client, XtPointer call)
{
	optionsPage *opp;
	switch (((XmAnyCallbackStruct *)call)->reason)
	{
		case XmCR_APPLY:
			*((int *)client) = OPTIONS_APPLY;
			for (opp = optionsPages; opp < &optionsPages[NELEM(optionsPages)]; opp++)
			{
				opp->pageCommit(opp);
			}
			break;
		case XmCR_CANCEL:
			*((int *)client) = OPTIONS_DISCARD;
			break;
	}
}
typedef struct _optionsMiscData {
	Widget speedWidget, fastWidget;
	Widget refreshWidget, cacheWidget, xsyncWidget, grabWidget;
	Widget smartWidget, warn48Widget;
} optionsMiscData;
static int
optionsMiscCreate(Widget parent, optionsPage *op)
{
	static optionsMiscData optionsMiscDatas;
	optionsMiscData *dp;
	Widget sep, w;
	XmString label;
	Arg arg[13];
	Cardinal i;
	op->pageData = dp = &optionsMiscDatas;
	XtVaSetValues(parent, XmNfractionBase, 100, XtNull);
	label = XmStringCreateLtoR((char *)"Emulation speed (%)", XmFONTLIST_DEFAULT_TAG);
	i = 0;
	XtSetArg(arg[i], XmNuserData,            op               ); i++;
	XtSetArg(arg[i], XmNorientation,         XmHORIZONTAL     ); i++;
	XtSetArg(arg[i], XmNprocessingDirection, XmMAX_ON_RIGHT   ); i++;
	XtSetArg(arg[i], XmNminimum,             10               ); i++;
	XtSetArg(arg[i], XmNmaximum,             200              ); i++;
	XtSetArg(arg[i], XmNscaleMultiple,       10               ); i++;
	XtSetArg(arg[i], XmNshowValue,           True             ); i++;
	XtSetArg(arg[i], XmNtitleString,         label            ); i++;
	XtSetArg(arg[i], XmNleftAttachment,      XmATTACH_FORM    ); i++;
	XtSetArg(arg[i], XmNrightAttachment,     XmATTACH_POSITION); i++;
	XtSetArg(arg[i], XmNrightPosition,       45               ); i++;
	XtSetArg(arg[i], XmNtopAttachment,       XmATTACH_FORM    ); i++;
	assert(i <= XtNumber(arg));
	w = XmCreateScale(parent, (char *)"scale", arg, i);
	XtManageChild(w);
	dp->speedWidget = w;
	XmStringFree(label);
	ToolhelpAddWidget(w, "Control emulation speed. The default\n"
						 "setting is 100, which causes the\n"
						 "emulation to never run faster than the\n"
						 "real machine. A higher value allows the\n"
						 "emulator to run faster, a lower one may\n"
						 "force it to run slower.");
	i = 0;
	XtSetArg(arg[i], XmNuserData,       op             ); i++;
	XtSetArg(arg[i], XmNleftAttachment, XmATTACH_FORM  ); i++;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_WIDGET); i++;
	XtSetArg(arg[i], XmNtopWidget,      w              ); i++;
	assert(i <= XtNumber(arg));
	w = XmCreateToggleButton(parent, (char *)"Fast mode", arg, i);
	XtManageChild(w);
	dp->fastWidget = w;
	ToolhelpAddWidget(w, "Special speed mode, which disables any\n"
						 "speed limit, turns sound emulation off\n"
						 "and uses a 1/10 refresh rate, so that\n"
						 "the emulator will run at the maximum\n"
						 "possible speed.");
	label = XmStringCreateLtoR((char *)"Refresh rate (1/n)", XmFONTLIST_DEFAULT_TAG);
	i = 0;
	XtSetArg(arg[i], XmNuserData,            op               ); i++;
	XtSetArg(arg[i], XmNorientation,         XmHORIZONTAL     ); i++;
	XtSetArg(arg[i], XmNprocessingDirection, XmMAX_ON_RIGHT   ); i++;
	XtSetArg(arg[i], XmNminimum,             1                ); i++;
	XtSetArg(arg[i], XmNmaximum,             50               ); i++;
	XtSetArg(arg[i], XmNscaleMultiple,       1                ); i++;
	XtSetArg(arg[i], XmNshowValue,           True             ); i++;
	XtSetArg(arg[i], XmNtitleString,         label            ); i++;
	XtSetArg(arg[i], XmNleftAttachment,      XmATTACH_POSITION); i++;
	XtSetArg(arg[i], XmNleftPosition,        50               ); i++;
	XtSetArg(arg[i], XmNrightAttachment,     XmATTACH_POSITION); i++;
	XtSetArg(arg[i], XmNrightPosition,       95               ); i++;
	XtSetArg(arg[i], XmNtopAttachment,       XmATTACH_FORM    ); i++;
	assert(i <= XtNumber(arg));
	w = XmCreateScale(parent, (char *)"scale", arg, i);
	XtManageChild(w);
	dp->refreshWidget = w;
	XmStringFree(label);
	ToolhelpAddWidget(w, "Control rate of frame update. The value\n"
						 "ranges from \"1/1\" (update 1/1 of the\n"
						 "frames of the real machine, that is 50\n"
						 "frames per second) to \"1/50\" (update 1\n"
						 "every 50 frames).");
	i = 0;
	XtSetArg(arg[i], XmNuserData,       op               ); i++;
	XtSetArg(arg[i], XmNleftAttachment, XmATTACH_POSITION); i++;
	XtSetArg(arg[i], XmNleftPosition,   50               ); i++;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_WIDGET  ); i++;
	XtSetArg(arg[i], XmNtopWidget,      w                ); i++;
	assert(i <= XtNumber(arg));
	w = XmCreateToggleButton(parent, (char *)"Video cache", arg, i);
	XtManageChild(w);
	dp->cacheWidget = w;
	ToolhelpAddWidget(w, "Video cache can speed up the emulation\n"
						 "when little graphics activity is going\n"
						 "on. It is especially useful when the\n"
						 "emulator is running across a network, as\n"
						 "it can reduce the amount of bandwidth\n"
						 "required.");
	i = 0;
	XtSetArg(arg[i], XmNuserData,       op               ); i++;
	XtSetArg(arg[i], XmNleftAttachment, XmATTACH_POSITION); i++;
	XtSetArg(arg[i], XmNleftPosition,   50               ); i++;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_WIDGET  ); i++;
	XtSetArg(arg[i], XmNtopWidget,      w                ); i++;
	assert(i <= XtNumber(arg));
	w = XmCreateToggleButton(parent, (char *)"XSync", arg, i);
	XtManageChild(w);
	dp->xsyncWidget = w;
	ToolhelpAddWidget(w, "Call the X11 function XSync before\n"
						 "updating the emulation window. This\n"
						 "might be necessary on low-end systems\n"
						 "to prevent it from stealing so many\n"
						 "system resources that it would become\n"
						 "impossible to interact with it.");
	i = 0;
	XtSetArg(arg[i], XmNuserData,       op               ); i++;
#ifndef XZX_KMOUSE
	XtSetArg(arg[i], XmNsensitive,      False            ); i++;
#endif
	XtSetArg(arg[i], XmNleftAttachment, XmATTACH_POSITION); i++;
	XtSetArg(arg[i], XmNleftPosition,   50               ); i++;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_WIDGET  ); i++;
	XtSetArg(arg[i], XmNtopWidget,      w                ); i++;
	assert(i <= XtNumber(arg));
	w = XmCreateToggleButton(parent, (char *)"Pointer grab", arg, i);
	XtManageChild(w);
	dp->grabWidget = w;
#ifdef XZX_KMOUSE
	ToolhelpAddWidget(w, "Actively grab control of the pointer.\n"
						 "Further pointer events are reported\n"
						 "only to the emulator. The pointer is\n"
						 "restricted to stay contained in that\n"
						 "window.");
#endif
	i = 0;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_WIDGET); i++;
	XtSetArg(arg[i], XmNtopWidget,      w              ); i++;
	XtSetArg(arg[i], XmNtopOffset,      15             ); i++;
	assert(i <= XtNumber(arg));
	sep = XmCreateSeparatorGadget(parent, (char *)"sep", arg, i);
	XtManageChild(sep);
	i = 0;
	XtSetArg(arg[i], XmNuserData,       op             ); i++;
	XtSetArg(arg[i], XmNsensitive,      False          ); i++;
	XtSetArg(arg[i], XmNleftAttachment, XmATTACH_FORM  ); i++;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_WIDGET); i++;
	XtSetArg(arg[i], XmNtopWidget,      sep            ); i++;
	assert(i <= XtNumber(arg));
	w = XmCreateToggleButtonGadget(parent, (char *)"Smart attach of media", arg, i);
	XtManageChild(w);
	dp->smartWidget = w;
	i = 0;
	XtSetArg(arg[i], XmNuserData,       op             ); i++;
	XtSetArg(arg[i], XmNleftAttachment, XmATTACH_FORM  ); i++;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_WIDGET); i++;
	XtSetArg(arg[i], XmNtopWidget,      w              ); i++;
	assert(i <= XtNumber(arg));
	w = XmCreateToggleButtonGadget(parent, (char *)"Warn if a 48k program accesses\nports from the 128K machinery", arg, i);
	XtManageChild(w);
	dp->warn48Widget = w;
	return 0;
}
static void
optionsMiscCallback(Widget w, XtPointer client, XtPointer call)
{
}
static void
optionsMiscUpdate(optionsPage *op)
{
	optionsMiscData *dp = (optionsMiscData *)(op->pageData);
	XmScaleSetValue(dp->speedWidget, CFGGET(speed));
	XmToggleButtonSetState(dp->fastWidget, (CFGGET(fastmode) & 1) != 0 ? True : False, False);
	XmScaleSetValue(dp->refreshWidget, CFGGET(refresh));
	XmToggleButtonSetState(dp->cacheWidget, CFGGET(cache) != 0 ? True : False, False);
	XmToggleButtonSetState(dp->xsyncWidget, CFGGET(xsync) != 0 ? True : False, False);
#ifdef XZX_KMOUSE
	XmToggleButtonSetState(dp->grabWidget, CFGGET(kmouseGrab) != 0 ? True : False, False);
#endif
	XmToggleButtonSetState(dp->warn48Widget, CFGGET(warn48) != 0 ? True : False, False);
}
static void
optionsMiscCommit(optionsPage *op)
{
	optionsMiscData *dp = (optionsMiscData *)(op->pageData);
	int val;
	XmScaleGetValue(dp->speedWidget, &val);
	val = ((val + 5) / 10) * 10;
	CFGSET(speed, val);
	EmulSetSpeed(val);
	CFGSET(fastmode, (CFGGET(fastmode) & ~1) | (XmToggleButtonGetState(dp->fastWidget) == True ? 1 : 0));
#ifdef REAL_AUDIO
	AudioControl(CFGGET(fastmode) == 0 ? AUC_ON : AUC_OFF);
#endif
	XmScaleGetValue(dp->refreshWidget, &val);
	CFGSET(refresh, val);
	CFGSET(cache, XmToggleButtonGetState(dp->cacheWidget) == True ? 1 : 0);
	CFGSET(xsync, XmToggleButtonGetState(dp->xsyncWidget) == True ? 1 : 0);
#ifdef XZX_KMOUSE
	CFGSET(kmouseGrab, XmToggleButtonGetState(dp->grabWidget) == True ? 1 : 0);
#endif
	CFGSET(warn48, XmToggleButtonGetState(dp->warn48Widget) == True ? 1 : 0);
}
enum { HW_SP482, HW_SP483, HW_SP128, HW_SP3, HW_PENTAGON, HW_SCORPION, HW_DIDAKTIK, HW_MAX_OPTIONS };
enum { HW_RESET, HW_IF1, HW_MF, HW_TRDOS, HW_MGT, HW_USPEECH, HW_KMOUSE, HW_MAX_BUTTONS };
typedef struct _optionsHardwareData {
	Widget modelWidget;
	unsigned int modelValue;
	struct {
		const char *label;
		Boolean sensitive;
		Widget widget;
	} modelOptions[HW_MAX_OPTIONS];
	struct {
		const char *label;
		Boolean sensitive;
		Widget widget;
	} hardwareButtons[HW_MAX_BUTTONS];
} optionsHardwareData;
static int
optionsHardwareCreate(Widget parent, optionsPage *op)
{
	static optionsHardwareData optionsHardwareDatas = {
		(Widget)0, 0,
		{
			{ "Spectrum 48K Issue 2", True,  (Widget)0 },
			{ "Spectrum 48K Issue 3", True,  (Widget)0 },
			{ "Spectrum 128K",        True,  (Widget)0 },
			{ "Spectrum +3",          True,  (Widget)0 },
#ifdef XZX_PENTAGON
			{ "Pentagon 128",         True,  (Widget)0 },
#else
			{ "Pentagon 128",         False, (Widget)0 },
#endif
#ifdef XZX_SCORPION
			{ "Scorpion 256",         True,  (Widget)0 },
#else
			{ "Scorpion 256",         False, (Widget)0 },
#endif
#ifdef XZX_DIDAKTIK
			{ "Didaktik-M",           True,  (Widget)0 },
#else
			{ "Didaktik-M",           False, (Widget)0 },
#endif
		},
		{
			{ "Reset upon change",    True,  (Widget)0 },
#ifdef XZX_IF1
			{ "Interface I",          True,  (Widget)0 },
#else
			{ "Interface I",          False, (Widget)0 },
#endif
#ifdef XZX_MF
			{ "Multiface",            True,  (Widget)0 },
#else
			{ "Multiface",            False, (Widget)0 },
#endif
#ifdef XZX_TRDOS
			{ "Beta 128",             True,  (Widget)0 },
#else
			{ "Beta 128",             False, (Widget)0 },
#endif
#ifdef XZX_MGT
			{ "+D",                   True,  (Widget)0 },
#else
			{ "+D",                   False, (Widget)0 },
#endif
			{ "Currah uSpeech",       False, (Widget)0 },
#ifdef XZX_KMOUSE
			{ "Kempston mouse",       True,  (Widget)0 },
#else
			{ "Kempston mouse",       False, (Widget)0 },
#endif
		}
	};
	optionsHardwareData *dp;
	Widget menu, cascade, sep, w;
	XmString label;
	Arg arg[7];
	Cardinal i;
	int pos, j;
	op->pageData = dp = &optionsHardwareDatas;
	menu = XmCreatePulldownMenu(parent, (char *)"pd", (ArgList)0, 0);
	label = XmStringCreateLtoR((char *)"Mode", XmFONTLIST_DEFAULT_TAG);
	i = 0;
	XtSetArg(arg[i], XmNsubMenuId,      menu         ); i++;
	XtSetArg(arg[i], XmNlabelString,    label        ); i++;
	XtSetArg(arg[i], XmNleftAttachment, XmATTACH_FORM); i++;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_FORM); i++;
	assert(i <= XtNumber(arg));
	cascade = XmCreateOptionMenu(parent, (char *)"om", arg, i);
	XmStringFree(label);
	for (j = HW_SP482; j < HW_MAX_OPTIONS; j++)
	{
		i = 0;
		XtSetArg(arg[i], XmNuserData,  op                           ); i++;
		XtSetArg(arg[i], XmNsensitive, dp->modelOptions[j].sensitive); i++;
		assert(i <= XtNumber(arg));
		w = XmCreatePushButton(menu, (char *)(dp->modelOptions[j].label), arg, i);
		XtManageChild(w);
		if (dp->modelOptions[j].sensitive == True)
		{
			XtAddCallback(w, XmNactivateCallback, optionsHardwareCallback, j);
		}
		dp->modelOptions[j].widget = w;
	}
	XtManageChild(cascade);
	dp->modelWidget = cascade;
	i = 0;
	XtSetArg(arg[i], XmNuserData,       op             ); i++;
	XtSetArg(arg[i], XmNset,            True           ); i++;
	XtSetArg(arg[i], XmNleftAttachment, XmATTACH_FORM  ); i++;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_WIDGET); i++;
	XtSetArg(arg[i], XmNtopWidget,      cascade        ); i++;
	assert(i <= XtNumber(arg));
	w = XmCreateToggleButton(parent, (char *)(dp->hardwareButtons[HW_RESET].label), arg, i);
	XtManageChild(w);
	dp->hardwareButtons[HW_RESET].widget = w;
	ToolhelpAddWidget(w, "Usually it is a good idea to reset the\n"
						 "computer when changes are made to the\n"
						 "emulated hardware, because of software\n"
						 "initialisation. By default, a hardware\n"
						 "change (except from the extra hardware)\n"
						 "causes a reset. If this behaviour is not\n"
						 "desired, it can be switched off. The\n"
						 "emulator will try to lift the current\n"
						 "state to the new environment as well as\n"
						 "possible.");
	i = 0;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_WIDGET); i++;
	XtSetArg(arg[i], XmNtopWidget,      w              ); i++;
	XtSetArg(arg[i], XmNtopOffset,      20             ); i++;
	assert(i <= XtNumber(arg));
	sep = XmCreateSeparatorGadget(parent, (char *)"sep", arg, i);
	XtManageChild(sep);
	XtVaSetValues(parent, XmNfractionBase, 2, XtNull);
	for (j = HW_IF1, pos = 0, w = sep; j < HW_MAX_BUTTONS; j++)
	{
		if (j == HW_USPEECH)
		{
			pos++;
			w = sep;
		}
		i = 0;
		XtSetArg(arg[i], XmNuserData,       op                              ); i++;
		XtSetArg(arg[i], XmNsensitive,      dp->hardwareButtons[j].sensitive); i++;
		XtSetArg(arg[i], XmNleftAttachment, XmATTACH_POSITION               ); i++;
		XtSetArg(arg[i], XmNleftPosition,   pos                             ); i++;
		XtSetArg(arg[i], XmNleftOffset,     5                               ); i++;
		XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_WIDGET                 ); i++;
		XtSetArg(arg[i], XmNtopWidget,      w                               ); i++;
		assert(i <= XtNumber(arg));
		w = XmCreateToggleButtonGadget(parent, (char *)(dp->hardwareButtons[j].label), arg, i);
		XtManageChild(w);
		XtAddCallback(w, XmNvalueChangedCallback, optionsHardwareCallback, HW_MAX_OPTIONS + j);
		dp->hardwareButtons[j].widget = w;
	}
	return 0;
}
static void
optionsHardwareCallback(Widget w, XtPointer client, XtPointer call)
{
	Boolean set = ((XmToggleButtonCallbackStruct *)call)->set != 0 ? True : False;
	Boolean s_if1, s_mf, s_trdos, s_mgt, s_uspeech, s_kmouse;
	Boolean v_if1, v_mf, v_trdos, v_mgt, v_uspeech, v_kmouse;
	optionsPage *op;
	optionsHardwareData *dp;
	XtVaGetValues(w, XmNuserData, &op, XtNull);
	dp = (optionsHardwareData *)(op->pageData);
	assert(dp != (optionsHardwareData *)0);
	s_if1 = XtIsSensitive(dp->hardwareButtons[HW_IF1].widget);
	s_mf = XtIsSensitive(dp->hardwareButtons[HW_MF].widget);
	s_trdos = XtIsSensitive(dp->hardwareButtons[HW_TRDOS].widget);
	s_mgt = XtIsSensitive(dp->hardwareButtons[HW_MGT].widget);
	s_uspeech = XtIsSensitive(dp->hardwareButtons[HW_USPEECH].widget);
	s_kmouse = XtIsSensitive(dp->hardwareButtons[HW_KMOUSE].widget);
	v_if1 = XmToggleButtonGadgetGetState(dp->hardwareButtons[HW_IF1].widget);
	v_mf = XmToggleButtonGadgetGetState(dp->hardwareButtons[HW_MF].widget);
	v_trdos = XmToggleButtonGadgetGetState(dp->hardwareButtons[HW_TRDOS].widget);
	v_mgt = XmToggleButtonGadgetGetState(dp->hardwareButtons[HW_MGT].widget);
	v_uspeech = XmToggleButtonGadgetGetState(dp->hardwareButtons[HW_USPEECH].widget);
	v_kmouse = XmToggleButtonGadgetGetState(dp->hardwareButtons[HW_KMOUSE].widget);
	switch ((int)client)
	{
		case HW_SP482:
		case HW_SP483:
		case HW_SP128:
		case HW_SP3:
		case HW_PENTAGON:
		case HW_SCORPION:
		case HW_DIDAKTIK:
			dp->modelValue = (unsigned int)client;
#ifdef XZX_IF1
			if (CFGGET(if1) >= 0)
			{
				s_if1 = dp->modelValue > HW_SP128 ? False : True;
			}
#endif
#ifdef XZX_MF
			s_mf = dp->modelValue > HW_SP3 ? False : True;
#endif
#ifdef XZX_TRDOS
			if (CFGGET(trdos) >= 0)
			{
				s_trdos = dp->modelValue > HW_SP128 ? False : True;
			}
#endif
#ifdef XZX_MGT
			if (CFGGET(mgt) >= 0)
			{
				s_mgt = dp->modelValue > HW_SP128 ? False : True;
			}
#endif
#ifdef XZX_KMOUSE
			s_kmouse = dp->modelValue == HW_PENTAGON || dp->modelValue == HW_SCORPION ? True : False;
#endif
			break;
		case HW_MAX_OPTIONS + HW_IF1:
			v_if1 = set;
			if (v_if1 == True)
			{
				v_mgt = False;
			}
			break;
		case HW_MAX_OPTIONS + HW_MF:
			v_mf = set;
			if (v_mf == True)
			{
				v_trdos = v_mgt = False;
			}
			break;
		case HW_MAX_OPTIONS + HW_TRDOS:
			v_trdos = set;
			if (v_trdos == True)
			{
				v_mf = v_mgt = False;
			}
			break;
		case HW_MAX_OPTIONS + HW_MGT:
			v_mgt = set;
			if (v_mgt == True)
			{
				v_if1 = v_mf = v_trdos = False;
			}
			break;
		case HW_MAX_OPTIONS + HW_USPEECH:
			v_uspeech = set;
			break;
		case HW_MAX_OPTIONS + HW_KMOUSE:
			v_kmouse = set;
			break;
	}
	XmToggleButtonGadgetSetState(dp->hardwareButtons[HW_IF1].widget, s_if1 == True ? v_if1 : False, False);
	XmToggleButtonGadgetSetState(dp->hardwareButtons[HW_MF].widget, s_mf == True ? v_mf : False, False);
	XmToggleButtonGadgetSetState(dp->hardwareButtons[HW_TRDOS].widget, s_trdos == True ? v_trdos : False, False);
	XmToggleButtonGadgetSetState(dp->hardwareButtons[HW_MGT].widget, s_mgt == True ? v_mgt : False, False);
	XmToggleButtonGadgetSetState(dp->hardwareButtons[HW_USPEECH].widget, s_uspeech == True ? v_uspeech : False, False);
	XmToggleButtonGadgetSetState(dp->hardwareButtons[HW_KMOUSE].widget, s_kmouse == True ? v_kmouse : False, False);
	XtSetSensitive(dp->hardwareButtons[HW_IF1].widget, s_if1);
	XtSetSensitive(dp->hardwareButtons[HW_MF].widget, s_mf);
	XtSetSensitive(dp->hardwareButtons[HW_TRDOS].widget, s_trdos);
	XtSetSensitive(dp->hardwareButtons[HW_MGT].widget, s_mgt);
	XtSetSensitive(dp->hardwareButtons[HW_USPEECH].widget, s_uspeech);
	XtSetSensitive(dp->hardwareButtons[HW_KMOUSE].widget, s_kmouse);
}
static void
optionsHardwareUpdate(optionsPage *op)
{
	optionsHardwareData *dp = (optionsHardwareData *)(op->pageData);
	int model = CFGGET(machine);
#if !defined(XZX_PENTAGON) && !defined(XZX_SCORPION)
	if (model > SP_3)
	{
		model += 2;
	}
#elif !defined(XZX_PENTAGON)
	if (model > SP_3)
	{
		model++;
	}
#elif !defined(XZX_SCORPION)
	if (model > SP_PENTAGON)
	{
		model++;
	}
#endif
	XtVaSetValues(dp->modelWidget, XmNmenuHistory, dp->modelOptions[model].widget, XtNull);
	dp->modelValue = model;
#ifdef XZX_IF1
	XmToggleButtonGadgetSetState(dp->hardwareButtons[HW_IF1].widget, CFGGET(if1) > 0 ? True : False, False);
	XtSetSensitive(dp->hardwareButtons[HW_IF1].widget, CFGGET(if1) >= 0 && CFGGET(machine) <= SP_128 ? True : False);
#endif
#ifdef XZX_MF
	XmToggleButtonGadgetSetState(dp->hardwareButtons[HW_MF].widget, CFGGET(mf1) > 0 ? True : False, False);
	XtSetSensitive(dp->hardwareButtons[HW_MF].widget, CFGGET(machine) <= SP_3 ? True : False);
#endif
#ifdef XZX_TRDOS
	XmToggleButtonGadgetSetState(dp->hardwareButtons[HW_TRDOS].widget, CFGGET(trdos) > 0 ? True : False, False);
	XtSetSensitive(dp->hardwareButtons[HW_TRDOS].widget, CFGGET(trdos) >= 0 && CFGGET(machine) <= SP_128 ? True : False);
#endif
#ifdef XZX_MGT
	XmToggleButtonGadgetSetState(dp->hardwareButtons[HW_MGT].widget, CFGGET(mgt) > 0 ? True : False, False);
	XtSetSensitive(dp->hardwareButtons[HW_MGT].widget, CFGGET(mgt) >= 0 && CFGGET(machine) <= SP_128 ? True : False);
#endif
#ifdef XZX_KMOUSE
	XmToggleButtonGadgetSetState(dp->hardwareButtons[HW_KMOUSE].widget, CFGGET(kmouse) > 0 ? True : False, False);
	XtSetSensitive(dp->hardwareButtons[HW_KMOUSE].widget, PENTAGON_OR_SCORPION ? True : False);
#endif
}
static void
optionsHardwareCommit(optionsPage *op)
{
	optionsHardwareData *dp = (optionsHardwareData *)(op->pageData);
	int model = dp->modelValue;
#ifndef XZX_SCORPION
	if (dp->modelValue > HW_SCORPION)
	{
		model--;
	}
#endif
#ifndef XZX_PENTAGON
	if (dp->modelValue > HW_PENTAGON)
	{
		model--;
	}
#endif
	if (model != CFGGET(machine))
	{
		MachineSelect(model, XmToggleButtonGetState(dp->hardwareButtons[HW_RESET].widget) == True ? 1 : 0);
	}
#ifdef XZX_IF1
	if (CFGGET(if1) >= 0)
	{
		CFGSET(if1, XmToggleButtonGadgetGetState(dp->hardwareButtons[HW_IF1].widget) == True ? 1 : 0);
		MenuFileCartridgeUpdate();
	}
#endif
#ifdef XZX_MF
	CFGSET(mf1, XmToggleButtonGadgetGetState(dp->hardwareButtons[HW_MF].widget) == True ? 1 : 0);
#endif
#ifdef XZX_TRDOS
	if (CFGGET(trdos) >= 0)
	{
		CFGSET(trdos, XmToggleButtonGadgetGetState(dp->hardwareButtons[HW_TRDOS].widget) == True ? 1 : 0);
	}
#endif
#ifdef XZX_MGT
	if (CFGGET(mgt) >= 0)
	{
		CFGSET(mgt, XmToggleButtonGadgetGetState(dp->hardwareButtons[HW_MGT].widget) == True ? 1 : 0);
	}
#endif
#ifdef XZX_KMOUSE
	CFGSET(kmouse, XmToggleButtonGadgetGetState(dp->hardwareButtons[HW_KMOUSE].widget) == True ? 1 : 0);
#endif
	MenuFileDiskUpdate();
}
enum { TAPE_QUICK, TAPE_FAST, TAPE_STOP, TAPE_NOISE, TAPE_MAX_BUTTONS };
typedef struct _optionsTapeData {
	struct {
		const char *label;
		Boolean sensitive;
		Widget widget;
	} buttons[TAPE_MAX_BUTTONS];
} optionsTapeData;
static int
optionsTapeCreate(Widget parent, optionsPage *op)
{
	static optionsTapeData optionsTapeDatas = {
		{
#ifndef NO_ED_TRAPS
			{ "Quick loading of blocks",   True,  (Widget)0 },
#else
			{ "Quick loading of blocks",   False, (Widget)0 },
#endif
			{ "Fast mode while loading",   True,  (Widget)0 },
			{ "Stop playback after block", True,  (Widget)0 },
#ifdef REAL_AUDIO
			{ "Noise while loading",       True,  (Widget)0 },
#else
			{ "Noise while loading",       False, (Widget)0 },
#endif
		}
	};
	optionsTapeData *dp;
	Widget w = (Widget)0;
	Arg arg[5];
	Cardinal i;
	unsigned int j;
	op->pageData = dp = &optionsTapeDatas;
	for (j = TAPE_QUICK; j <= TAPE_NOISE; j++)
	{
		i = 0;
		XtSetArg(arg[i], XmNuserData,          op                      ); i++;
		XtSetArg(arg[i], XmNsensitive,         dp->buttons[j].sensitive); i++;
		XtSetArg(arg[i], XmNleftAttachment,    XmATTACH_FORM           ); i++;
		if (w == (Widget)0)
		{
			XtSetArg(arg[i], XmNtopAttachment, XmATTACH_FORM           ); i++;
		}
		else
		{
			XtSetArg(arg[i], XmNtopAttachment, XmATTACH_WIDGET         ); i++;
			XtSetArg(arg[i], XmNtopWidget,     w                       ); i++;
		}
		assert(i <= XtNumber(arg));
		w = XmCreateToggleButtonGadget(parent, (char *)(dp->buttons[j].label), arg, i);
		XtManageChild(w);
		if (dp->buttons[j].sensitive == True)
		{
			XtAddCallback(w, XmNvalueChangedCallback, optionsTapeCallback, j);
		}
		dp->buttons[j].widget = w;
	}
	return 0;
}
static void
optionsTapeCallback(Widget w, XtPointer client, XtPointer call)
{
	optionsPage *op;
	optionsTapeData *dp;
	XtVaGetValues(w, XmNuserData, &op, XtNull);
	dp = (optionsTapeData *)(op->pageData);
	assert(dp != (optionsTapeData *)0);
	switch ((int)client)
	{
		case TAPE_QUICK:
		case TAPE_STOP:
		case TAPE_NOISE:
			break;
		case TAPE_FAST:
#ifdef REAL_AUDIO
			{
				Boolean state = ((XmToggleButtonCallbackStruct *)call)->set != 0 ? True : False;
				Widget button = dp->buttons[TAPE_NOISE].widget;
				if (state == True)
				{
					XmToggleButtonGadgetSetState(button, False, False);
				}
				XtSetSensitive(button, !state);
			}
#endif
			break;
	}
}
static void
optionsTapeUpdate(optionsPage *op)
{
	optionsTapeData *dp = (optionsTapeData *)(op->pageData);
	Boolean faststate;
#ifndef NO_ED_TRAPS
	XmToggleButtonGadgetSetState(dp->buttons[TAPE_QUICK].widget, CFGGET(tapeQuickload) == 1 ? True : False, False);
#endif
	faststate = CFGGET(tapeFastload) == 1 ? True : False;
	XmToggleButtonGadgetSetState(dp->buttons[TAPE_FAST].widget, faststate, False);
#ifdef REAL_AUDIO
	XmToggleButtonGadgetSetState(dp->buttons[TAPE_NOISE].widget, CFGGET(tapeNoise) == 1 ? True : False, False);
	XtSetSensitive(dp->buttons[TAPE_NOISE].widget, !faststate);
#endif
}
static void
optionsTapeCommit(optionsPage *op)
{
	optionsTapeData *dp = (optionsTapeData *)(op->pageData);
	unsigned int tapestate = TapeGetState();
	int savestate;
#ifndef NO_ED_TRAPS
	if (XmToggleButtonGadgetGetState(dp->buttons[TAPE_QUICK].widget) == True)
	{
		CFGSET(tapeQuickload, 1);
		tapestate |= TS_QUICKLOAD;
	}
	else
	{
		CFGSET(tapeQuickload, 0);
		tapestate &= ~TS_QUICKLOAD;
	}
#endif
	savestate = CFGGET(tapeFastload);
	if (XmToggleButtonGadgetGetState(dp->buttons[TAPE_FAST].widget) == True)
	{
		CFGSET(tapeFastload, 1);
		tapestate |= TS_FASTLOAD;
	}
	else
	{
		CFGSET(tapeFastload, 0);
		tapestate &= ~TS_FASTLOAD;
	}
	if (XmToggleButtonGadgetGetState(dp->buttons[TAPE_STOP].widget) == True)
	{
		CFGSET(tapeAutostop, 1);
		tapestate |= TS_AUTOSTOP;
	}
	else
	{
		CFGSET(tapeAutostop, 0);
		tapestate &= ~TS_AUTOSTOP;
	}
#ifdef REAL_AUDIO
	if (XmToggleButtonGadgetGetState(dp->buttons[TAPE_NOISE].widget) == True)
	{
		CFGSET(tapeNoise, 1);
		tapestate |= TS_NOISE;
	}
	else
	{
		CFGSET(tapeNoise, 0);
		tapestate &= ~TS_NOISE;
	}
#endif
	if (savestate != CFGGET(tapeFastload))
	{
		if ((tapestate & (TS_PLAY | TS_FASTLOAD)) == (TS_PLAY | TS_FASTLOAD))
		{
			CFGSET(fastmode, CFGGET(fastmode) | 2);
		}
		else
		{
			CFGSET(fastmode, CFGGET(fastmode) & ~2);
		}
#ifdef REAL_AUDIO
		AudioControl(CFGGET(fastmode) == 0 ? AUC_ON : AUC_OFF);
#endif
	}
	TapeSetState(tapestate);
}
#ifdef AUDIO
enum { SOUND_SP, SOUND_AY, SOUND_USPEECH, SOUND_MAX_SCALES };
enum { SOUND_MONO, SOUND_ACB, SOUND_ABC, SOUND_MAX_RADIOS };
typedef struct _optionsSoundData {
	struct {
		const char *label;
		Boolean sensitive;
		Widget widget;
		const char *toolhelp;
	} scales[SOUND_MAX_SCALES];
	struct {
		const char *label;
		Widget widget;
	} radios[SOUND_MAX_RADIOS];
} optionsSoundData;
static int
optionsSoundCreate(Widget parent, optionsPage *op)
{
	static optionsSoundData optionsSoundDatas = {
		{
			{ "SP", True,  (Widget)0, "Speaker volume" },
#ifdef REAL_AUDIO
			{ "AY", True,  (Widget)0, "AY-8912 volume" },
#else
			{ "AY", False, (Widget)0, (const char *)0  },
#endif
			{ "US", False, (Widget)0, (const char *)0  }
		},
		{
			{ "Mono", (Widget)0 },
			{ "ACB",  (Widget)0 },
			{ "ABC",  (Widget)0 }
		}
	};
	optionsSoundData *dp;
	Widget form, radio, sep, w;
	Arg arg[13];
	Cardinal i;
	unsigned int j;
	op->pageData = dp = &optionsSoundDatas;
	XtVaSetValues(parent, XmNfractionBase, 100, XtNull);
	for (j = SOUND_SP, form = (Widget)0; j < SOUND_MAX_SCALES; j++)
	{
		i = 0;
		if (form == (Widget)0)
		{
			XtSetArg(arg[i], XmNleftAttachment, XmATTACH_FORM    ); i++;
			XtSetArg(arg[i], XmNleftOffset,     10               ); i++;
		}
		else
		{
			XtSetArg(arg[i], XmNleftAttachment, XmATTACH_WIDGET  ); i++;
			XtSetArg(arg[i], XmNleftWidget,     form             ); i++;
			XtSetArg(arg[i], XmNleftOffset,     20               ); i++;
		}
		XtSetArg(arg[i], XmNtopAttachment,      XmATTACH_FORM    ); i++;
		XtSetArg(arg[i], XmNbottomAttachment,   XmATTACH_POSITION); i++;
		XtSetArg(arg[i], XmNbottomPosition,     65               ); i++;
		assert(i <= XtNumber(arg));
		form = XmCreateForm(parent, (char *)"form", arg, i);
		i = 0;
		XtSetArg(arg[i], XmNrightAttachment,  XmATTACH_FORM); i++;
		XtSetArg(arg[i], XmNtopAttachment,    XmATTACH_FORM); i++;
		assert(i <= XtNumber(arg));
		w = XmCreateLabelGadget(form, (char *)dp->scales[j].label, arg, i);
		XtManageChild(w);
		i = 0;
		XtSetArg(arg[i], XmNuserData,            op                     ); i++;
		XtSetArg(arg[i], XmNsensitive,           dp->scales[j].sensitive); i++;
		XtSetArg(arg[i], XmNorientation,         XmVERTICAL             ); i++;
		XtSetArg(arg[i], XmNprocessingDirection, XmMAX_ON_TOP           ); i++;
		XtSetArg(arg[i], XmNminimum,             0                      ); i++;
		XtSetArg(arg[i], XmNmaximum,             100                    ); i++;
		XtSetArg(arg[i], XmNscaleMultiple,       10                     ); i++;
		XtSetArg(arg[i], XmNshowValue,           False                  ); i++;
		XtSetArg(arg[i], XmNleftAttachment,      XmATTACH_FORM          ); i++;
		XtSetArg(arg[i], XmNrightAttachment,     XmATTACH_FORM          ); i++;
		XtSetArg(arg[i], XmNtopAttachment,       XmATTACH_WIDGET        ); i++;
		XtSetArg(arg[i], XmNtopWidget,           w                      ); i++;
		XtSetArg(arg[i], XmNbottomAttachment,    XmATTACH_FORM          ); i++;
		assert(i <= XtNumber(arg));
		w = XmCreateScale(form, (char *)"scale", arg, i);
		XtManageChild(w);
		dp->scales[j].widget = w;
		XmScaleSetTicks(w, 10, 1, 0, 4, 2, 0);
		if (dp->scales[j].sensitive == True)
		{
			ToolhelpAddWidget(w, dp->scales[j].toolhelp);
		}
		XtManageChild(form);
	}
	i = 0;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_FORM  ); i++;
	XtSetArg(arg[i], XmNleftAttachment, XmATTACH_WIDGET); i++;
	XtSetArg(arg[i], XmNleftWidget,     form           ); i++;
	XtSetArg(arg[i], XmNleftOffset,     25             ); i++;
	assert(i <= XtNumber(arg));
	sep = XmCreateSeparatorGadget(parent, (char *)"sep", arg, i);
	XtManageChild(sep);
	i = 0;
#ifndef REAL_AUDIO
	XtSetArg(arg[i], XmNsensitive,      False                ); i++;
#endif
	XtSetArg(arg[i], XmNalignment,      XmALIGNMENT_BEGINNING); i++;
	XtSetArg(arg[i], XmNleftAttachment, XmATTACH_WIDGET      ); i++;
	XtSetArg(arg[i], XmNleftWidget,     sep                  ); i++;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_FORM        ); i++;
	assert(i <= XtNumber(arg));
	w = XmCreateLabelGadget(parent, (char *)"AY mode:", arg, i);
	XtManageChild(w);
	i = 0;
#ifndef REAL_AUDIO
	XtSetArg(arg[i], XmNsensitive,      False          ); i++;
#endif
	XtSetArg(arg[i], XmNleftAttachment, XmATTACH_WIDGET); i++;
	XtSetArg(arg[i], XmNleftWidget,     sep            ); i++;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_WIDGET); i++;
	XtSetArg(arg[i], XmNtopWidget,      w              ); i++;
	assert(i <= XtNumber(arg));
	radio = XmCreateRadioBox(parent, (char *)"rb", arg, i);
	for (j = SOUND_MONO; j < SOUND_MAX_RADIOS; j++)
	{
		w = XmCreateToggleButtonGadget(radio, (char *)(dp->radios[j].label), (ArgList)0, 0);
		XtManageChild(w);
		dp->radios[j].widget = w;
	}
	XmToggleButtonGadgetSetState(dp->radios[SOUND_MONO].widget, True, True);
	XtManageChild(radio);
	return 0;
}
static void
optionsSoundCallback(Widget w, XtPointer client, XtPointer call)
{
}
static void
optionsSoundUpdate(optionsPage *op)
{
	optionsSoundData *dp = (optionsSoundData *)(op->pageData);
	unsigned int val;
#ifndef REAL_AUDIO
	val = CFGGET(audioVolume);
#else
	val = AudioGetVolume(AUV_SP) * 25;
#endif
	XmScaleSetValue(dp->scales[SOUND_SP].widget, (int)val);
#ifdef REAL_AUDIO
	val = AudioGetVolume(AUV_AY) * 25;
	XmScaleSetValue(dp->scales[SOUND_AY].widget, (int)val);
	XmToggleButtonGadgetSetState(dp->radios[Audio.stereo == 0 ? SOUND_MONO : CFGGET(audioStereo)].widget, True, True);
	XtSetSensitive(XtParent(dp->radios[SOUND_MONO].widget), Audio.stereo == 0 ? False : True);
#endif
}
static void
optionsSoundCommit(optionsPage *op)
{
	optionsSoundData *dp = (optionsSoundData *)(op->pageData);
	unsigned int val;
	XmScaleGetValue(dp->scales[SOUND_SP].widget, (int *)(&val));
#ifndef REAL_AUDIO
	CFGSET(audioVolume, val);
	CFGSET(audio, val > 0 ? 1 : 0);
#else
	val /= 25;
	AudioSetVolume(AUV_SP, val);
	CFGSET(audio, val > 0 ? 1 : 0);
	XmScaleGetValue(dp->scales[SOUND_AY].widget, (int *)(&val));
	val /= 25;
	AudioSetVolume(AUV_AY, val);
	CFGSET(audio, val > 0 ? 1 : CFGGET(audio));
	if (XtIsSensitive(XtParent(dp->radios[SOUND_MONO].widget)) == True)
	{
		int j;
		for (j = SOUND_MONO; j < SOUND_MAX_RADIOS; j++)
		{
			if (XmToggleButtonGadgetGetState(dp->radios[j].widget) == True)
			{
				break;
			}
		}
		assert(j < SOUND_MAX_RADIOS);
		CFGSET(audioStereo, j);
	}
#endif
}
#endif
enum { JOY_UP, JOY_DOWN, JOY_LEFT, JOY_RIGHT, JOY_FIRE, JOY_MAX_KEYS };
enum { JOY_SINCLAIR1, JOY_SINCLAIR2, JOY_CURSOR, JOY_KEMPSTON, JOY_MAX_OPTIONS };
enum { JOY_TOLERANCE, JOY_AUTOFIRE, JOY_JUGGLE, JOY_MAX_SPINS };
typedef struct _optionsJoystickData {
	Widget keyWidget;
	struct {
		const char *label;
		Widget widget;
	} keys[JOY_MAX_KEYS];
	Widget joyWidget;
	Widget joyType;
	struct {
		const char *label;
		Widget widget;
	} joyOptions[JOY_MAX_OPTIONS];
	struct {
		const char *label;
		Widget widget;
	} joySpins[JOY_MAX_SPINS];
} optionsJoystickData;
static int
optionsJoystickCreate(Widget parent, optionsPage *op)
{
	static optionsJoystickData optionsJoystickDatas = {
		(Widget)0,
		{
			{ "Up",   (Widget)0 },
			{ "Down", (Widget)0 },
			{ "Left", (Widget)0 },
			{ "Right",(Widget)0 },
			{ "Fire", (Widget)0 },
		},
		(Widget)0,
		(Widget)0,
		{
			{ "Sinclair 1", (Widget)0 },
			{ "Sinclair 2", (Widget)0 },
			{ "Cursor",     (Widget)0 },
			{ "Kempston",   (Widget)0 },
		},
		{
			{ "Tolerance",  (Widget)0 },
			{ "Autofire",   (Widget)0 },
			{ "L-R Juggle", (Widget)0 },
		}
	};
	optionsJoystickData *dp;
	Widget form, menu, cascade, tf, w;
	XmString label;
	Arg arg[15];
	Cardinal i;
	unsigned int j;
	op->pageData = dp = &optionsJoystickDatas;
	XtVaSetValues(parent, XmNfractionBase, 100, XtNull);
	i = 0;
	XtSetArg(arg[i], XmNuserData,       op           ); i++;
	XtSetArg(arg[i], XmNset,            True         ); i++;
	XtSetArg(arg[i], XmNindicatorType,  XmONE_OF_MANY); i++;
	XtSetArg(arg[i], XmNleftAttachment, XmATTACH_FORM); i++;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_FORM); i++;
	assert(i <= XtNumber(arg));
	w = XmCreateToggleButtonGadget(parent, (char *)"Keyboard", arg, i);
	XtManageChild(w);
	XtAddCallback(w, XmNvalueChangedCallback, optionsJoystickCallback, JOY_MAX_KEYS);
	dp->keyWidget = w;
	for (j = JOY_UP; j < JOY_MAX_KEYS; j++)
	{
		i = 0;
		XtSetArg(arg[i], XmNleftAttachment,  XmATTACH_FORM    ); i++;
		XtSetArg(arg[i], XmNrightAttachment, XmATTACH_POSITION); i++;
		XtSetArg(arg[i], XmNrightPosition,   45               ); i++;
		XtSetArg(arg[i], XmNtopAttachment,   XmATTACH_WIDGET  ); i++;
		XtSetArg(arg[i], XmNtopWidget,       w                ); i++;
		assert(i <= XtNumber(arg));
		form = XmCreateForm(parent, (char *)"form", arg, i);
		i = 0;
		XtSetArg(arg[i], XmNleftAttachment,   XmATTACH_FORM); i++;
		XtSetArg(arg[i], XmNtopAttachment,    XmATTACH_FORM); i++;
		XtSetArg(arg[i], XmNbottomAttachment, XmATTACH_FORM); i++;
		assert(i <= XtNumber(arg));
		w = XmCreateLabelGadget(form, (char *)dp->keys[j].label, arg, i);
		XtManageChild(w);
		i = 0;
		XtSetArg(arg[i], XmNuserData,              op               ); i++;
		XtSetArg(arg[i], XmNcolumns,               8                ); i++;
		XtSetArg(arg[i], XmNcursorPositionVisible, False            ); i++;
		XtSetArg(arg[i], XmNleftAttachment,        XmATTACH_POSITION); i++;
		XtSetArg(arg[i], XmNleftPosition,          30               ); i++;
		XtSetArg(arg[i], XmNtopAttachment,         XmATTACH_FORM    ); i++;
		XtSetArg(arg[i], XmNbottomAttachment,      XmATTACH_FORM    ); i++;
		assert(i <= XtNumber(arg));
		w = XmCreateTextField(form, (char *)"tf", arg, i);
		XtManageChild(w);
		XtAddCallback(w, XmNfocusCallback, optionsJoystickCallback, j);
		XtAddCallback(w, XmNlosingFocusCallback, optionsJoystickCallback, j);
		dp->keys[j].widget = w;
		XtManageChild(form);
	}
	i = 0;
	XtSetArg(arg[i], XmNuserData,       op               ); i++;
	XtSetArg(arg[i], XmNsensitive,      False            ); i++;
	XtSetArg(arg[i], XmNset,            False            ); i++;
	XtSetArg(arg[i], XmNindicatorType,  XmONE_OF_MANY    ); i++;
	XtSetArg(arg[i], XmNleftAttachment, XmATTACH_POSITION); i++;
	XtSetArg(arg[i], XmNleftPosition,   50               ); i++;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_FORM    ); i++;
	assert(i <= XtNumber(arg));
	w = XmCreateToggleButtonGadget(parent, (char *)"Joystick", arg, i);
	XtManageChild(w);
	XtAddCallback(w, XmNvalueChangedCallback, optionsJoystickCallback, JOY_MAX_KEYS + JOY_MAX_OPTIONS);
	dp->joyWidget = w;
	menu = XmCreatePulldownMenu(parent, (char *)"pd", (ArgList)0, 0);
	label = XmStringCreateLtoR((char *)"Type", XmFONTLIST_DEFAULT_TAG);
	i = 0;
#ifndef JOY
	XtSetArg(arg[i], XmNsensitive,      False            ); i++;
#endif
	XtSetArg(arg[i], XmNsubMenuId,      menu             ); i++;
	XtSetArg(arg[i], XmNlabelString,    label            ); i++;
	XtSetArg(arg[i], XmNleftAttachment, XmATTACH_POSITION); i++;
	XtSetArg(arg[i], XmNleftPosition,   50               ); i++;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_WIDGET  ); i++;
	XtSetArg(arg[i], XmNtopWidget,      w                ); i++;
	assert(i <= XtNumber(arg));
	cascade = XmCreateOptionMenu(parent, (char *)"om", arg, i);
	XmStringFree(label);
	for (j = JOY_SINCLAIR1; j < JOY_MAX_OPTIONS; j++)
	{
		i = 0;
		XtSetArg(arg[i], XmNuserData,  op               ); i++;
		XtSetArg(arg[i], XmNsensitive, j == JOY_KEMPSTON); i++;
		assert(i <= XtNumber(arg));
		w = XmCreatePushButton(menu, (char *)(dp->joyOptions[j].label), arg, i);
		XtManageChild(w);
		dp->joyOptions[j].widget = w;
	}
	XtManageChild(cascade);
	dp->joyType = cascade;
	for (j = JOY_TOLERANCE, w = cascade; j < JOY_MAX_SPINS; j++)
	{
		i = 0;
#ifndef JOY
		XtSetArg(arg[i], XmNsensitive,       False            ); i++;
#endif
		XtSetArg(arg[i], XmNleftAttachment,  XmATTACH_POSITION); i++;
		XtSetArg(arg[i], XmNleftPosition,    50               ); i++;
		XtSetArg(arg[i], XmNrightAttachment, XmATTACH_FORM    ); i++;
		XtSetArg(arg[i], XmNtopAttachment,   XmATTACH_WIDGET  ); i++;
		XtSetArg(arg[i], XmNtopWidget,       w                ); i++;
		assert(i <= XtNumber(arg));
		form = XmCreateForm(parent, (char *)"form", arg, i);
		i = 0;
		XtSetArg(arg[i], XmNleftAttachment,   XmATTACH_FORM); i++;
		XtSetArg(arg[i], XmNtopAttachment,    XmATTACH_FORM); i++;
		XtSetArg(arg[i], XmNbottomAttachment, XmATTACH_FORM); i++;
		assert(i <= XtNumber(arg));
		w = XmCreateLabelGadget(form, (char *)dp->joySpins[j].label, arg, i);
		XtManageChild(w);
		i = 0;
		XtSetArg(arg[i], XmNuserData,         op               ); i++;
		XtSetArg(arg[i], XmNarrowLayout,      XmARROWS_END     ); i++;
		XtSetArg(arg[i], XmNspinBoxChildType, XmNUMERIC        ); i++;
		XtSetArg(arg[i], XmNpositionType,     XmPOSITION_VALUE ); i++;
		XtSetArg(arg[i], XmNcolumns,          3                ); i++;
		XtSetArg(arg[i], XmNminimumValue,     0                ); i++;
		XtSetArg(arg[i], XmNmaximumValue,     50               ); i++;
		XtSetArg(arg[i], XmNincrementValue,   1                ); i++;
		XtSetArg(arg[i], XmNwrap,             False            ); i++;
		XtSetArg(arg[i], XmNeditable,         False            ); i++;
		XtSetArg(arg[i], XmNshadowThickness,  0                ); i++;
		XtSetArg(arg[i], XmNleftAttachment,   XmATTACH_POSITION); i++;
		XtSetArg(arg[i], XmNleftPosition,     50               ); i++;
		XtSetArg(arg[i], XmNtopAttachment,    XmATTACH_FORM    ); i++;
		XtSetArg(arg[i], XmNbottomAttachment, XmATTACH_FORM    ); i++;
		assert(i <= XtNumber(arg));
		w = XmCreateSimpleSpinBox(form, (char *)"sb", arg, i);
		XtManageChild(w);
		dp->joySpins[j].widget = w;
		XtVaGetValues(w, XmNtextField, &tf, XtNull);
		assert(tf != (Widget)0);
		XtVaSetValues(tf, XmNcursorPositionVisible, False, XtNull);
		XtManageChild(form);
	}
	return 0;
}
static void
optionsJoystickCallback(Widget w, XtPointer client, XtPointer call)
{
	optionsPage *op;
	optionsJoystickData *dp;
	KeySym key;
	char *str;
	XtVaGetValues(w, XmNuserData, &op, XtNull);
	dp = (optionsJoystickData *)(op->pageData);
	assert(dp != (optionsJoystickData *)0);
	switch ((int)client)
	{
		case JOY_UP:
		case JOY_DOWN:
		case JOY_LEFT:
		case JOY_RIGHT:
		case JOY_FIRE:
			if (((XmAnyCallbackStruct *)call)->reason == XmCR_FOCUS)
			{
				XtVaSetValues(w, XmNcursorPositionVisible, True, XtNull);
			}
			else
			{
				XtVaSetValues(w, XmNcursorPositionVisible, False, XtNull);
				XmTextFieldClearSelection(w, CurrentTime);
				str = XmTextFieldGetString(w);
				key = XStringToKeysym(str);
				XtFree(str);
				if (key == NoSymbol)
				{
					(void)XBell(XtDisplayOfObject(w), 5);
				}
			}
			break;
		case JOY_MAX_KEYS:
#ifdef JOY
			if (XtIsSensitive(dp->joyWidget) == True)
			{
				XmToggleButtonGadgetSetState(dp->joyWidget, False, False);
			}
			else
#endif
			{
				XmToggleButtonGadgetSetState(w, True, False);
			}
			break;
#ifdef JOY
		case JOY_MAX_KEYS + JOY_MAX_OPTIONS:
			XmToggleButtonGadgetSetState(dp->keyWidget, False, False);
			break;
#endif
	}
}
static void
optionsJoystickUpdate(optionsPage *op)
{
	optionsJoystickData *dp = (optionsJoystickData *)(op->pageData);
#ifdef JOY
	Boolean sensitive;
	int tolerance, autofire, juggle;
#endif
	unsigned int j;
	for (j = JOY_UP; j < JOY_MAX_KEYS; j++)
	{
		XmTextFieldSetString(dp->keys[j].widget, XKeysymToString(CFGGET(joyKeys[j])));
	}
#ifdef JOY
	sensitive = CFGGET(joy) >= 0 ? True : False;
	tolerance = CFGGET(joyTolerance);
	autofire = CFGGET(joyAutofire);
	juggle = CFGGET(joyJuggle);
	XtSetSensitive(dp->joyWidget, sensitive);
	XtSetSensitive(dp->joyType, sensitive);
	XtSetSensitive(XtParent(dp->joySpins[JOY_TOLERANCE].widget), sensitive);
	XtSetSensitive(XtParent(dp->joySpins[JOY_AUTOFIRE].widget), sensitive == True && autofire > 0);
	XtSetSensitive(XtParent(dp->joySpins[JOY_JUGGLE].widget), sensitive == True && juggle > 0);
	if (sensitive == True)
	{
		XmToggleButtonGadgetSetState(dp->joyWidget, CFGGET(joy) > 0 ? True : False, True);
		XtVaSetValues(dp->joySpins[JOY_TOLERANCE].widget, XmNposition, tolerance, XtNull);
		if (autofire > 0)
		{
			XtVaSetValues(dp->joySpins[JOY_AUTOFIRE].widget, XmNposition, autofire, XtNull);
		}
		if (juggle > 0)
		{
			XtVaSetValues(dp->joySpins[JOY_JUGGLE].widget, XmNposition, juggle, XtNull);
		}
	}
#endif
}
static void
optionsJoystickCommit(optionsPage *op)
{
	optionsJoystickData *dp = (optionsJoystickData *)(op->pageData);
	KeySym key;
	char *str;
	unsigned int j;
	for (j = JOY_UP; j < JOY_MAX_KEYS; j++)
	{
		str = XmTextFieldGetString(dp->keys[j].widget);
		key = XStringToKeysym(str);
		XtFree(str);
		if (key != NoSymbol)
		{
			CFGSET(joyKeys[j], key);
		}
	}
#ifdef JOY
	if (CFGGET(joy) >= 0)
	{
		CFGSET(joy, XmToggleButtonGadgetGetState(dp->joyWidget) == True ? 1 : 0);
		XtVaGetValues(dp->joySpins[JOY_TOLERANCE].widget, XmNposition, &CFGGET(joyTolerance), XtNull);
		if (CFGGET(joyAutofire) > 0)
		{
			XtVaGetValues(dp->joySpins[JOY_AUTOFIRE].widget, XmNposition, &CFGGET(joyAutofire), XtNull);
		}
		if (CFGGET(joyJuggle) > 0)
		{
			XtVaGetValues(dp->joySpins[JOY_JUGGLE].widget, XmNposition, &CFGGET(joyJuggle), XtNull);
		}
	}
#endif
}
#ifdef XZX_IF1
enum { IF1_NONE, IF1_STDOUT, IF1_RS232, IF1_PRT, IF1_MAX_OPTIONS };
enum { IF1_CRLF, IF1_STRCR, IF1_MAX_BUTTONS };
typedef struct _optionsIf1Data {
	Widget rs232Widget;
	struct {
		const char *label;
		Boolean sensitive;
		Widget widget;
	} rs232Options[IF1_MAX_OPTIONS];
	struct {
		const char *label;
		Widget widget;
		const char *toolhelp;
	} buttons[IF1_MAX_BUTTONS];
} optionsIf1Data;
static int
optionsIf1Create(Widget parent, optionsPage *op)
{
	static optionsIf1Data optionsIf1Datas = {
		(Widget)0,
		{
			{ "none",              False, (Widget)0 },
			{ "stdin/stdout",      True,  (Widget)0 },
			{ "serial port",       False, (Widget)0 },
#ifdef xIF1_PRINTER
			{ "printer emulation", True,  (Widget)0 },
#else
			{ "printer emulation", False, (Widget)0 },
#endif
		},
		{
			{ "Translate newline",
			  (Widget)0,
			  "Converts line feeds to carriage returns\n"
			  "on RS232 input, and vice versa on RS232\n"
			  "output. This is necessary if textual data\n"
			  "is transfered from UNIX to the emulator."
			},
			{ "Strip carriage return",
			  (Widget)0,
			  "Strip carriage returns on RS232 output.\n"
			  "This is only useful when listing BASIC\n"
			  "programs to the Interface I \"T\" channel."
			}
		}
	};
	optionsIf1Data *dp;
	Widget menu, cascade, w;
	XmString label;
	Arg arg[4];
	Cardinal i;
	int j;
	op->pageData = dp = &optionsIf1Datas;
	menu = XmCreatePulldownMenu(parent, (char *)"pd", (ArgList)0, 0);
	label = XmStringCreateLtoR((char *)"RS232 mode", XmFONTLIST_DEFAULT_TAG);
	i = 0;
	XtSetArg(arg[i], XmNsubMenuId,      menu         ); i++;
	XtSetArg(arg[i], XmNlabelString,    label        ); i++;
	XtSetArg(arg[i], XmNleftAttachment, XmATTACH_FORM); i++;
	XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_FORM); i++;
	assert(i <= XtNumber(arg));
	cascade = XmCreateOptionMenu(parent, (char *)"om", arg, i);
	XmStringFree(label);
	for (j = IF1_NONE; j < IF1_MAX_OPTIONS; j++)
	{
		i = 0;
		XtSetArg(arg[i], XmNuserData,  op                           ); i++;
		XtSetArg(arg[i], XmNsensitive, dp->rs232Options[j].sensitive); i++;
		assert(i <= XtNumber(arg));
		w = XmCreatePushButton(menu, (char *)(dp->rs232Options[j].label), arg, i);
		XtManageChild(w);
		if (dp->rs232Options[j].sensitive == True)
		{
			XtAddCallback(w, XmNactivateCallback, optionsIf1Callback, j);
		}
		dp->rs232Options[j].widget = w;
	}
	XtManageChild(cascade);
	dp->rs232Widget = cascade;
	for (j = IF1_CRLF, w = cascade; j < IF1_MAX_BUTTONS; j++)
	{
		i = 0;
		XtSetArg(arg[i], XmNuserData,       op             ); i++;
		XtSetArg(arg[i], XmNleftAttachment, XmATTACH_FORM  ); i++;
		XtSetArg(arg[i], XmNtopAttachment,  XmATTACH_WIDGET); i++;
		XtSetArg(arg[i], XmNtopWidget,      w              ); i++;
		assert(i <= XtNumber(arg));
		w = XmCreateToggleButton(parent, (char *)(dp->buttons[j].label), arg, i);
		XtManageChild(w);
		dp->buttons[j].widget = w;
		ToolhelpAddWidget(w, dp->buttons[j].toolhelp);
	}
	return 0;
}
static void
optionsIf1Callback(Widget w, XtPointer client, XtPointer call)
{
	optionsPage *op;
	optionsIf1Data *dp;
	XtVaGetValues(w, XmNuserData, &op, XtNull);
	dp = (optionsIf1Data *)(op->pageData);
	assert(dp != (optionsIf1Data *)0);
	switch ((int)client)
	{
		case IF1_NONE:
		case IF1_STDOUT:
		case IF1_RS232:
		case IF1_PRT:
			break;
	}
}
static void
optionsIf1Update(optionsPage *op)
{
	optionsIf1Data *dp = (optionsIf1Data *)(op->pageData);
	Boolean sensitive = CFGGET(if1) >= 0 ? True : False;
	Boolean state;
	XtSetSensitive(dp->rs232Widget, sensitive);
	state = sensitive == True ? CFGGET(if1Crlf) == 1 ? True : False : False;
	XmToggleButtonSetState(dp->buttons[IF1_CRLF].widget, state, False);
	XtSetSensitive(dp->buttons[IF1_CRLF].widget, sensitive);
	state = sensitive == True ? CFGGET(if1Strcr) == 1 ? True : False : False;
	XmToggleButtonSetState(dp->buttons[IF1_STRCR].widget, state, False);
	XtSetSensitive(dp->buttons[IF1_STRCR].widget, sensitive);
}
static void
optionsIf1Commit(optionsPage *op)
{
	optionsIf1Data *dp = (optionsIf1Data *)(op->pageData);
	if (XtIsSensitive(dp->buttons[IF1_CRLF].widget) == True)
	{
		CFGSET(if1Crlf, XmToggleButtonGetState(dp->buttons[IF1_CRLF].widget) == True ? 1 : 0);
	}
	if (XtIsSensitive(dp->buttons[IF1_STRCR].widget) == True)
	{
		CFGSET(if1Strcr, XmToggleButtonGetState(dp->buttons[IF1_STRCR].widget) == True ? 1 : 0);
	}
}
#endif
#ifdef PRINTER
static int optionsPrinterCreate(Widget parent, optionsPage *op) {return 0;}
static void optionsPrinterCallback(Widget w, XtPointer client, XtPointer call) {}
static void optionsPrinterUpdate(optionsPage *op) {}
static void optionsPrinterCommit(optionsPage *op) {}
#endif
#ifdef DEBUG
typedef struct _optionsDebugData {
	const char *label;
	Boolean sensitive;
	Widget widget;
	unsigned int mask;
	unsigned int value;
} optionsDebugData;
static int
optionsDebugCreate(Widget parent, optionsPage *op)
{
	static optionsDebugData optionsDebugDatas[] = {
		{ "Z80 mnemonics",         True,  (Widget)0, D_MNEMO,     0 },
		{ "Z80 registers",         True,  (Widget)0, D_CPU,       0 },
		{ "Z80 in/out",            True,  (Widget)0, D_IO,        0 },
		{ "Z80 interrupts",        True,  (Widget)0, D_INTERRUPT, 0 },
		{ "Z80 illegal opcodes",   True,  (Widget)0, D_ILLOPS,    0 },
		{ "uPD765A emulation",     True,  (Widget)0, D_UPD765A,   0 },
#if defined(XZX_TRDOS) || defined(XZX_MGT)
		{ "WD17XX emulation",      True,  (Widget)0, D_WD17XX,    0 },
#else
		{ "WD17XX emulation",      False, (Widget)0, D_WD17XX,    0 },
#endif
#ifdef XZX_IF1
		{ "Interface I emulation", True,  (Widget)0, D_IF1,       0 },
#else
		{ "Interface I emulation", False, (Widget)0, D_IF1,       0 },
#endif
#ifdef XZX_TRDOS
		{ "Beta 128 emulation",    True,  (Widget)0, D_TRDOS,     0 },
#else
		{ "Beta 128 emulation",    False, (Widget)0, D_TRDOS,     0 },
#endif
#ifdef XZX_MGT
		{ "+D emulation",          True,  (Widget)0, D_MGT,       0 },
#else
		{ "+D emulation",          False, (Widget)0, D_MGT,       0 },
#endif
#ifdef AUDIO
		{ "Audio emulation",       True,  (Widget)0, D_AUDIO,     0 },
#else
		{ "Audio emulation",       False, (Widget)0, D_AUDIO,     0 },
#endif
#ifdef PRINTER
		{ "Printer emulation",     True,  (Widget)0, D_PRINTER,   0 },
#else
		{ "Printer emulation",     False, (Widget)0, D_PRINTER,   0 },
#endif
		{ "Tape operations",       True,  (Widget)0, D_TAPE,      0 },
		{ "Snapshot operations",   True,  (Widget)0, D_SNAP,      0 },
#ifdef JOY
		{ "Joystick operations",   True,  (Widget)0, D_JOY,       0 },
#else
		{ "Joystick operations",   False, (Widget)0, D_JOY,       0 },
#endif
		{ "X11 releated things",   True,  (Widget)0, D_X11,       0 },
		{ (const char *)0,         False, (Widget)0, 0,           0 }
	};
	optionsDebugData *dp;
	Widget rowcol;
	Arg arg[5];
	Cardinal i;
	unsigned int idx;
	op->pageData = dp = &optionsDebugDatas[0];
	i = 0;
	XtSetArg(arg[i], XmNnumColumns,     2            ); i++;
	XtSetArg(arg[i], XmNorientation,    XmVERTICAL   ); i++;
	XtSetArg(arg[i], XmNpacking,        XmPACK_COLUMN); i++;
	XtSetArg(arg[i], XmNmarginHeight,   5            ); i++;
	XtSetArg(arg[i], XmNmarginWidth,    5            ); i++;
	assert(i <= XtNumber(arg));
	rowcol = XmCreateRowColumn(parent, (char *)"rowcol", arg, i);
	XtManageChild(rowcol);
	for (idx = 0; dp->label != (const char *)0; dp++, idx++)
	{
		i = 0;
		XtSetArg(arg[i], XmNuserData,  op           ); i++;
		XtSetArg(arg[i], XmNsensitive, dp->sensitive); i++;
		assert(i <= XtNumber(arg));
		dp->widget = XmCreateToggleButtonGadget(rowcol, (char *)(dp->label), arg, i);
		XtManageChild(dp->widget);
		if (dp->sensitive == True)
		{
			XtAddCallback(dp->widget, XmNvalueChangedCallback, optionsDebugCallback, idx);
		}
	}
	return 0;
}
static void
optionsDebugCallback(Widget w, XtPointer client, XtPointer call)
{
	optionsPage *op;
	optionsDebugData *dp;
	XtVaGetValues(w, XmNuserData, &op, XtNull);
	dp = (optionsDebugData *)(op->pageData);
	assert(dp != (optionsDebugData *)0);
	dp += (unsigned int)client;
	if (((XmToggleButtonCallbackStruct *)call)->set != 0)
	{
		dp->value |= dp->mask;
	}
	else
	{
		dp->value &= ~(dp->mask);
	}
}
static void
optionsDebugUpdate(optionsPage *op)
{
	optionsDebugData *dp = (optionsDebugData *)(op->pageData);
	unsigned int debug = CFGGET(debug);
	for (; dp->label != (const char *)0; dp++)
	{
		if (dp->sensitive == True)
		{
			dp->value = debug & dp->mask;
			XmToggleButtonGadgetSetState(dp->widget, dp->value != 0 ? True : False, False);
		}
	}
}
static void
optionsDebugCommit(optionsPage *op)
{
	optionsDebugData *dp = (optionsDebugData *)(op->pageData);
	unsigned int debug = 0;
	for (; dp->label != (const char *)0; dp++)
	{
		if (dp->sensitive == True)
		{
			debug |= dp->value;
		}
	}
	CFGSET(debug, debug);
}
#endif
