/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
static char rcsid[] = "$Id: ui.c,v 1.9 2000/11/26 19:25:11 erik Rel $";
#include <Xm/Xm.h>
#include <Xm/DialogS.h>
#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
#include <Xm/List.h>
#include <Xm/MessageB.h>
#include <Xm/SelectioB.h>
#include <Xm/Text.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#ifdef XZX_KMOUSE
#include "resource.h"
#endif
#include "util.h"
#ifdef AUDIO
#include "audio.h"
#endif
#ifdef XZX_KMOUSE
#include "kmouse.h"
#endif
#include "main.h"
#include "xwin.h"
#include "ui.h"
typedef struct _fsData {
	char *fileName;
	int fileValid;
	int *fileTypes;
} fsData;
#define XtNull 							(XtPointer)0
#define XtAddCallback(w,cb,f,val)		XtAddCallback(w,cb,f,(XtPointer)val)
static void uiDialogCallback(Widget, XtPointer, XtPointer);
static void uiFileSelectorReadDirectory(Widget, XtPointer);
#ifdef HAVE_QSORT
static int uiFileSelectorCompare(const char **, const char **);
#endif
static void uiFileSelectorCallback(Widget, XtPointer, XtPointer);
static void uiFixiateListWidthCallback(Widget, XtPointer, XtPointer);
static int uiActive;
void
UiInit(int copyright)
{
	uiActive = 0;
	if (copyright != 0)
	{
		UiShowAbout();
	}
}
int
UiEnter(void)
{
	if (uiActive++ == 0)
	{
#ifdef REAL_AUDIO
		AudioControl(AUC_SUSPEND);
#endif
#ifdef XZX_KMOUSE
		if (CFGGET(kmouse) > 0)
		{
			KmouseOff();
		}
#endif
	}
	return uiActive;
}
int
UiLeave(void)
{
	uiActive--;
	assert(uiActive >= 0);
	if (uiActive == 0)
	{
#ifdef REAL_AUDIO
		AudioControl(AUC_RESUME);
#endif
#ifdef XZX_KMOUSE
		if (CFGGET(kmouse) > 0)
		{
			KmouseOn();
		}
#endif
	}
	return uiActive;
}
int
UiIsActive(void)
{
	return uiActive;
}
Widget
UiCreateFrame(Widget parent, const char *label)
{
	Widget frame, w;
	Arg arg[3];
	Cardinal i;
	i = 0;
	XtSetArg(arg[i], XmNshadowType,   XmSHADOW_ETCHED_IN); i++;
	XtSetArg(arg[i], XmNmarginHeight, 5                 ); i++;
	XtSetArg(arg[i], XmNmarginWidth,  5                 ); i++;
	assert(i <= XtNumber(arg));
	frame = XmCreateFrame(parent, (char *)"frame", arg, i);
	XtManageChild(frame);
	i = 0;
	XtSetArg(arg[i], XmNframeChildType,         XmFRAME_TITLE_CHILD); i++;
	XtSetArg(arg[i], XmNchildVerticalAlignment, XmALIGNMENT_CENTER ); i++;
	assert(i <= XtNumber(arg));
	w = XmCreateLabel(frame, (char *)label, arg, i);
	XtManageChild(w);
	return frame;
}
Widget
UiCreateDialog(int *answer)
{
	Widget dialog;
	XmString title, okLabel;
	Arg arg[7];
	Cardinal i;
	title = XmStringCreateLtoR((char *)"XZX Dialog", XmFONTLIST_DEFAULT_TAG);
	okLabel = XmStringCreateLtoR((char *)"Done", XmFONTLIST_DEFAULT_TAG);
	i = 0;
	XtSetArg(arg[i], XmNnoResize,            True                           ); i++;
	XtSetArg(arg[i], XmNautoUnmanage,        False                          ); i++;
	XtSetArg(arg[i], XmNdeleteResponse,      XmDO_NOTHING                   ); i++;
	XtSetArg(arg[i], XmNdialogStyle,         XmDIALOG_FULL_APPLICATION_MODAL); i++;
	XtSetArg(arg[i], XmNdialogTitle,         title                          ); i++;
	XtSetArg(arg[i], XmNokLabelString,       okLabel                        ); i++;
	XtSetArg(arg[i], XtNkeyboardFocusPolicy, XmPOINTER                      ); i++;
	assert(i <= XtNumber(arg));
	dialog = XmCreatePromptDialog(XwinDatas.topLevel, (char *)"options", arg, i);
	XmStringFree(title);
	XmStringFree(okLabel);
	XtUnmanageChild(XmSelectionBoxGetChild(dialog, XmDIALOG_COMMAND_TEXT));
	XtUnmanageChild(XmSelectionBoxGetChild(dialog, XmDIALOG_PROMPT_LABEL));
	XtUnmanageChild(XmSelectionBoxGetChild(dialog, XmDIALOG_CANCEL_BUTTON));
	XtUnmanageChild(XmSelectionBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));
	XtAddCallback(dialog, XmNokCallback, uiDialogCallback, answer);
	return dialog;
}
static void
uiDialogCallback(Widget w, XtPointer client, XtPointer call)
{
	switch (((XmAnyCallbackStruct *)call)->reason)
	{
		case XmCR_OK:
			*((int *)client) = DIALOG_OK;
			break;
		case XmCR_CANCEL:
			*((int *)client) = DIALOG_CANCEL;
			break;
	}
}
void
UiShowDialog(Widget dialog, int *answer)
{
	(void)UiEnter();
	XtManageChild(dialog);
	XtPopup(XtParent(dialog), XtGrabNone);
	*answer = DIALOG_NONE;
	while (*answer == DIALOG_NONE)
	{
		XtAppProcessEvent(XtWidgetToApplicationContext(dialog), XtIMAll);
	}
	XtUnmanageChild(dialog);
	XtPopdown(XtParent(dialog));
	(void)XSync(XtDisplay(XwinDatas.topLevel), 0);
	XmUpdateDisplay(XwinDatas.topLevel);
	(void)UiLeave();
}
void
UiShowAbout(void)
{
	char buf[256];
	(void)strcpy(buf, "          ");
	(void)strcat(buf, Version);
	(void)strcat(buf, "\n\nZX Spectrum 48/128/+3 emulator"
					  "\n  Pentagon/Scorpion emulator"
					  "\n       Didaktik emulator"
					  "\n\n    1995-2000 Erik Kunze"
					  "\n    1993-1994 Des Herriott");
	(void)UiShowMessage(DIALOG_MSG, buf, "Dismiss", (const char *)0, DIALOG_OK);
}
int
UiShowMessage(int type, const char *msgText, const char *okLabel, const char *cancelLabel, int defChoice)
{
	static Widget dialog = (Widget)0;
	static int answer;
	static XmString defOkLabel = (XmString)0, defCancellabel = (XmString)0;
	XmString str;
	if (dialog == (Widget)0)
	{
		Arg arg[5];
		Cardinal i;
		str = XmStringCreateLtoR((char *)"XZX Dialog", XmFONTLIST_DEFAULT_TAG);
		i = 0;
		XtSetArg(arg[i], XmNnoResize,       True                           ); i++;
		XtSetArg(arg[i], XmNautoUnmanage,   False                          ); i++;
		XtSetArg(arg[i], XmNdeleteResponse, XmDO_NOTHING                   ); i++;
		XtSetArg(arg[i], XmNdialogStyle,    XmDIALOG_FULL_APPLICATION_MODAL); i++;
		XtSetArg(arg[i], XmNdialogTitle,    str                            ); i++;
		assert(i <= XtNumber(arg));
		dialog = XmCreateMessageDialog(XwinDatas.topLevel, (char *)"dialog", arg, i);
		XmStringFree(str);
		XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));
		XtAddCallback(dialog, XmNokCallback, uiDialogCallback, &answer);
		XtAddCallback(dialog, XmNcancelCallback, uiDialogCallback, &answer);
		XtVaGetValues(dialog,
					  XmNokLabelString, &defOkLabel,
					  XmNcancelLabelString, &defCancellabel,
					  XtNull);
	}
	switch (type)
	{
		case DIALOG_ERROR:
			type = XmDIALOG_ERROR;
			break;
		case DIALOG_INFO:
			type = XmDIALOG_INFORMATION;
			break;
		case DIALOG_MSG:
			type = XmDIALOG_MESSAGE;
			break;
		case DIALOG_ASK:
			type = XmDIALOG_QUESTION;
			break;
		case DIALOG_WARN:
			type = XmDIALOG_WARNING;
			break;
		case DIALOG_WORK:
			type = XmDIALOG_WORKING;
			break;
	}
	XtVaSetValues(dialog, XmNdialogType, type, XtNull);
	if (type == XmDIALOG_QUESTION)
	{
		XtManageChild(XmMessageBoxGetChild(dialog, XmDIALOG_CANCEL_BUTTON));
	}
	else
	{
		XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_CANCEL_BUTTON));
	}
	str = XmStringCreateLtoR((char *)msgText, XmFONTLIST_DEFAULT_TAG);
	XtVaSetValues(dialog, XmNmessageString, str, XtNull);
	XmStringFree(str);
	if (okLabel != (const char *)0)
	{
		str = XmStringCreateLtoR((char *)okLabel, XmFONTLIST_DEFAULT_TAG);
		XtVaSetValues(dialog, XmNokLabelString, str, XtNull);
		XmStringFree(str);
	}
	else
	{
		XtVaSetValues(dialog, XmNokLabelString, defOkLabel, XtNull);
	}
	if (cancelLabel != (const char *)0)
	{
		str = XmStringCreateLtoR((char *)cancelLabel, XmFONTLIST_DEFAULT_TAG);
		XtVaSetValues(dialog, XmNcancelLabelString, str, XtNull);
		XmStringFree(str);
	}
	else
	{
		XtVaSetValues(dialog, XmNcancelLabelString, defCancellabel, XtNull);
	}
	if (defChoice == DIALOG_OK)
	{
		XtVaSetValues(dialog,
					  XmNdefaultButtonType, XmDIALOG_OK_BUTTON,
					  XmNinitialFocus, XmMessageBoxGetChild(dialog, XmDIALOG_OK_BUTTON),
					  XtNull);
	}
	else
	{
		XtVaSetValues(dialog,
					  XmNdefaultButtonType, XmDIALOG_CANCEL_BUTTON,
					  XmNinitialFocus, XmMessageBoxGetChild(dialog, XmDIALOG_CANCEL_BUTTON),
					  XtNull);
	}
	UiShowDialog(dialog, &answer);
	return answer;
}
int
UiShowTextFile(const char *fname, XtCallbackProc callback, XtPointer client)
{
	Widget dialog, form, scrolledText;
	Arg arg[8];
	Cardinal i;
	FILE *fp;
	struct stat sbuf;
	char *title, *text = (char *)0;
	if ((fp = Fopen((char *)fname, "rb")) == (FILE *)0)
	{
		return errno;
	}
	if ((fstat(fileno(fp), &sbuf) < 0)
		|| (text = malloc((size_t)(sbuf.st_size + 1))) == (char *)0
		|| fread(text, 1, (size_t)sbuf.st_size, fp) != (size_t)sbuf.st_size)
	{
		(void)fclose(fp);
		if (text != (char *)0)
		{
			free(text);
		}
		return errno == 0 ? -1 : errno;
	}
	(void)fclose(fp);
	text[sbuf.st_size] = '\0';
	if ((title = malloc(sizeof("XZX Browser: ") + strlen(fname) + 1)) == (char *)0)
	{
		free(text);
		return errno;
	}
	strcat(strcpy(title, "XZX Browser: "), fname);
	i = 0;
	XtSetArg(arg[i], XmNdeleteResponse, XmDESTROY); i++;
	assert(i <= XtNumber(arg));
	dialog = XmCreateDialogShell(XwinDatas.topLevel, title, arg, i);
	if (callback != (XtCallbackProc)0)
	{
		XtAddCallback(dialog, XmNdestroyCallback, callback, client);
	}
	free(title);
	i = 0;
	XtSetArg(arg[i], XmNmarginHeight, 5); i++;
	XtSetArg(arg[i], XmNmarginWidth,  5); i++;
	assert(i <= XtNumber(arg));
	form = XmCreateForm(dialog, (char *)"form", arg, i);
	i = 0;
	XtSetArg(arg[i], XmNrows,                  15               ); i++;
	XtSetArg(arg[i], XmNcolumns,               80               ); i++;
	XtSetArg(arg[i], XmNeditable,              False            ); i++;
	XtSetArg(arg[i], XmNeditMode,              XmMULTI_LINE_EDIT); i++;
	XtSetArg(arg[i], XmNcursorPositionVisible, False            ); i++;
	XtSetArg(arg[i], XmNvalue,                 text             ); i++;
	XtSetArg(arg[i], XmNscrollVertical,        True             ); i++;
	XtSetArg(arg[i], XmNscrollHorizontal,      False            ); i++;
	assert(i <= XtNumber(arg));
	scrolledText = XmCreateScrolledText(form, (char *)"st", arg, i);
	free(text);
	XtVaSetValues(XtParent(scrolledText),
				  XmNleftAttachment, XmATTACH_FORM,
				  XmNrightAttachment, XmATTACH_FORM,
				  XmNtopAttachment, XmATTACH_FORM,
				  XmNbottomAttachment, XmATTACH_FORM,
				  XtNull);
	XtManageChild(scrolledText);
	XtManageChild(form);
	XtManageChild(dialog);
	XtPopup(dialog, XtGrabNone);
	return 0;
}
char *
UiFileSelector(const char *title, int valid, int *types)
{
	static Widget dialog = (Widget)0;
	static fsData fsDatas;
	XmString str;
	(void)UiEnter();
	fsDatas.fileName = (char *)-1;
	fsDatas.fileValid = valid;
	fsDatas.fileTypes = types;
	if (dialog == (Widget)0)
	{
		Arg arg[6];
		Cardinal i;
		i = 0;
		XtSetArg(arg[i], XmNuserData,            &fsDatas                       ); i++;
		XtSetArg(arg[i], XmNautoUnmanage,        False                          ); i++;
		XtSetArg(arg[i], XmNdeleteResponse,      XmDO_NOTHING                   ); i++;
		XtSetArg(arg[i], XmNdialogStyle,         XmDIALOG_FULL_APPLICATION_MODAL); i++;
		XtSetArg(arg[i], XtNkeyboardFocusPolicy, XmPOINTER                      ); i++;
		XtSetArg(arg[i], XmNfileSearchProc,      uiFileSelectorReadDirectory    ); i++;
		assert(i <= XtNumber(arg));
		dialog = XmCreateFileSelectionDialog(XwinDatas.topLevel, (char *)"fs", arg, i);
		XtUnmanageChild(XmFileSelectionBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));
		XtAddCallback(dialog, XmNokCallback, uiFileSelectorCallback, &fsDatas);
		XtAddCallback(dialog, XmNcancelCallback, uiFileSelectorCallback, &fsDatas);
	}
	else
	{
		XtVaGetValues(dialog, XmNdirectory, &str, XtNull);
		XmFileSelectionDoSearch(dialog, str);
	}
	str = XmStringCreateLtoR((char *)title, XmFONTLIST_DEFAULT_TAG);
	XtVaSetValues(dialog, XmNdialogTitle, str, XtNull);
	XmStringFree(str);
	XtManageChild(dialog);
	XtPopup(XtParent(dialog), XtGrabNone);
	while (fsDatas.fileName == (char *)-1)
	{
		XtAppProcessEvent(XtWidgetToApplicationContext(dialog), XtIMAll);
	}
	XtUnmanageChild(dialog);
	XtPopdown(XtParent(dialog));
	(void)XSync(XtDisplay(XwinDatas.topLevel), 0);
	XmUpdateDisplay(XwinDatas.topLevel);
	(void)UiLeave();
	return fsDatas.fileName;
}
static void
uiFileSelectorReadDirectory(Widget w, XtPointer searchData)
{
	fsData *fsDatas;
	DIR *dir;
	struct dirent *ent;
	struct stat buf;
	char *saveDir, *scanDir, **filenames, *p;
	int cnt, max, *type, i;
	if (XmStringGetLtoR(((XmFileSelectionBoxCallbackStruct *)searchData)->dir, XmFONTLIST_DEFAULT_TAG, &scanDir) == False)
	{
		return;
	}
	XtVaGetValues(w, XmNuserData, &fsDatas, XtNull);
	assert(fsDatas != (fsData *)0);
	if ((saveDir = getcwd((char *)0, MAXPATHLEN)) == (char *)0)
	{
		Msg(M_FATAL, "getcwd failed");
	}
	if (chdir(scanDir) < 0)
	{
		XtFree(scanDir);
		free(saveDir);
		return;
	}
	if ((dir = opendir(scanDir)) == (DIR *)0)
	{
		Msg(M_PERR, "couldn't open directory <%s>", scanDir);
		XtFree(scanDir);
		(void)chdir(saveDir);
		free(saveDir);
		return;
	}
	cnt = 0;
	max = 256;
	filenames = Malloc((size_t)max * sizeof(char *), "uiFileSelectorReadDirectory");
	while ((ent = readdir(dir)) > (struct dirent *)0)
	{
		if (ent->d_name[0] == '.' && ent->d_name[1] != '.')
		{
			continue;
		}
		(void)stat(ent->d_name, &buf);
#ifdef VIRTUAL_DISK_IMAGES
		if (S_ISREG(buf.st_mode) || (S_ISDIR(buf.st_mode) && GetFileType(ent->d_name) == FT_TRD))
#else
		if (S_ISREG(buf.st_mode))
#endif
		{
			p = strrchr(ent->d_name, '.');
			for (type = fsDatas->fileTypes; *type != FT_END; type++)
			{
				if (*type == FT_ANY || (p != (char *)0 && (strcasecmp(p + 1, FtExt[*type])) == 0))
				{
					filenames[cnt] = Strdup(ent->d_name, "uiFileSelectorReadDirectory");
					if (++cnt == max)
					{
						max += 256;
						if ((filenames = realloc(filenames, (size_t)max * sizeof(char *))) == (char **)0)
						{
							Msg(M_FATAL, "realloc failed in uiFileSelectorReadDirectory()");
						}
					}
				}
			}
		}
	}
	(void)closedir(dir);
	(void)chdir(saveDir);
	free(saveDir);
	if (cnt > 0)
	{
		XmString *names;
		char *file;
#ifndef HAVE_QSORT
		int j;
		for (i = cnt - 1; i >= 0; i--)
		{
			for (j = 1; j <= i; j++)
			{
				if (strcasecmp(filenames[j - 1], filenames[j]) > 0)
				{
					p = filenames[j - 1];
					filenames[j - 1] = filenames[j];
					filenames[j] = p;
				}
			}
		}
#else
		qsort(filenames, (unsigned int)cnt, sizeof(filenames[0]), (int (*)(const void *, const void *))uiFileSelectorCompare);
#endif
		names = (XmString *)XtMalloc((size_t)cnt * sizeof(XmString *));
		file = XtMalloc(MAXPATHLEN);
		(void)strcpy(file, scanDir);
		p = file + strlen(file);
		for (i = 0; i < cnt; i++)
		{
			(void)strcpy(p, filenames[i]);
			names[i] = XmStringCreateLtoR(file, XmFONTLIST_DEFAULT_TAG);
			free(filenames[i]);
		}
		XtFree(file);
		XtVaSetValues(w,
					  XmNfileListItems, names,
					  XmNfileListItemCount, cnt,
					  XmNlistUpdated, True,
					  XtNull);
		while (cnt)
		{
			XmStringFree(names[--cnt]);
		}
		XtFree((char *)names);
	}
	else
	{
		XtVaSetValues(w,
					  XmNfileListItems, XtNull,
					  XmNfileListItemCount, 0,
					  XmNlistUpdated, True,
					  XtNull);
	}
	free(filenames);
	XtFree(scanDir);
}
#ifdef HAVE_QSORT
static int
uiFileSelectorCompare(const char **s1, const char **s2)
{
	return (strcasecmp(*s1, *s2));
}
#endif
static void
uiFileSelectorCallback(Widget w, XtPointer client, XtPointer call)
{
	XmFileSelectionBoxCallbackStruct *cbs = (XmFileSelectionBoxCallbackStruct *)call;
	char *file;
	fsData *fsDatas;
	struct stat buf;
	if (cbs->reason == XmCR_CANCEL)
	{
		*((char **)client) = (char *)0;
		return;
	}
	if (XmStringGetLtoR(cbs->value, XmFONTLIST_DEFAULT_TAG, &file) == False)
	{
		return;
	}
	XtVaGetValues(w, XmNuserData, &fsDatas, XtNull);
	assert(fsDatas != (fsData *)0);
	if (*file != '/')
	{
		char *dir, *newfile;
		if (XmStringGetLtoR(cbs->dir, XmFONTLIST_DEFAULT_TAG, &dir) == True)
		{
			newfile = XtMalloc(MAXPATHLEN);
			(void)strcat(strcpy(newfile, dir), file);
			XtFree(file);
			XtFree(dir);
			file = newfile;
		}
	}
	if (stat(file, &buf) == 0)
	{
		if (S_ISREG(buf.st_mode))
		{
			if (fsDatas->fileValid != 0
				|| UiShowMessage(DIALOG_ASK, "Overwrite file?", (const char *)0, (const char *)0, DIALOG_CANCEL) == DIALOG_OK)
			{
				goto finish;
			}
		}
		else if (S_ISDIR(buf.st_mode))
		{
			XmString str;
#ifdef VIRTUAL_DISK_IMAGES
			if (GetFileType(file) == FT_TRD && fsDatas->fileValid != 0)
			{
				goto finish;
			}
#endif
			str = XmStringCreateLtoR(file, XmFONTLIST_DEFAULT_TAG);
			XmFileSelectionDoSearch(w, str);
			XmStringFree(str);
		}
	}
	else if (errno == ENOENT && fsDatas->fileValid == 0)
	{
	  finish:
		*((void **)client) = Strdup(file, "uiFileSelectorCallback");
		XtPopdown(XtParent(w));
	}
	XtFree(file);
}
int
UiBrowseList(const char *title, char **items, int maxItems, int curItem)
{
	static Widget dialog = (Widget)0, list;
	static int answer;
	XmString str, *listItems;
	int cnt;
	if (dialog == (Widget)0)
	{
		XmString okLabel;
		Arg arg[7];
		Cardinal i;
		str = XmStringCreateLtoR((char *)"XZX Browser", XmFONTLIST_DEFAULT_TAG);
		okLabel = XmStringCreateLtoR((char *)"Apply", XmFONTLIST_DEFAULT_TAG);
		i = 0;
		XtSetArg(arg[i], XmNnoResize,            True                           ); i++;
		XtSetArg(arg[i], XmNautoUnmanage,        False                          ); i++;
		XtSetArg(arg[i], XmNdeleteResponse,      XmDO_NOTHING                   ); i++;
		XtSetArg(arg[i], XmNdialogStyle,         XmDIALOG_FULL_APPLICATION_MODAL); i++;
		XtSetArg(arg[i], XtNkeyboardFocusPolicy, XmPOINTER                      ); i++;
		XtSetArg(arg[i], XmNdialogTitle,         str                            ); i++;
		XtSetArg(arg[i], XmNokLabelString,       okLabel                        ); i++;
		assert(i <= XtNumber(arg));
		dialog = XmCreateSelectionDialog(XwinDatas.topLevel, (char *)"browser", arg, i);
		list = XmSelectionBoxGetChild(dialog, XmDIALOG_LIST);
		XmStringFree(str);
		XmStringFree(okLabel);
		XtUnmanageChild(XmSelectionBoxGetChild(dialog, XmDIALOG_APPLY_BUTTON));
		XtUnmanageChild(XmSelectionBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));
		XtUnmanageChild(XmSelectionBoxGetChild(dialog, XmDIALOG_SELECTION_LABEL));
		XtUnmanageChild(XmSelectionBoxGetChild(dialog, XmDIALOG_TEXT));
		XtAddCallback(dialog, XmNokCallback, uiDialogCallback, &answer);
		XtAddCallback(dialog, XmNcancelCallback, uiDialogCallback, &answer);
	}
	str = XmStringCreateLtoR((char *)title, XmFONTLIST_DEFAULT_TAG);
	listItems = (XmString *)XtMalloc((size_t)maxItems * sizeof(XmString *));
	for (cnt = 0; cnt < maxItems; cnt++)
	{
		listItems[cnt] = XmStringCreateLtoR(items[cnt], XmFONTLIST_DEFAULT_TAG);
	}
	XtVaSetValues(dialog,
				  XmNlistLabelString, str,
				  XmNlistItems, listItems,
				  XmNlistItemCount, maxItems,
				  XtNull);
	XmStringFree(str);
	curItem++;
	XmListSelectPos(list, curItem, False);
	XmListSetPos(list, curItem);
	curItem = -1;
	while (--cnt >= 0)
	{
		XmStringFree(listItems[cnt]);
	}
	XtFree((char *)listItems);
	UiShowDialog(dialog, &answer);
	if (answer == DIALOG_OK)
	{
		int *posList, posCnt;
		if (XmListGetSelectedPos(list, &posList, &posCnt) == True && posCnt == 1)
		{
			curItem = *posList - 1;
		}
		XtFree((char *)posList);
	}
	return curItem;
}
void
UiFixiateListWidth(Widget shell, Widget list, int width)
{
	XmString item;
	char *str;
	str = XtCalloc(sizeof(char), (Cardinal)(width + 1));
	for (--width; width >= 0; width--)
	{
		str[width] = 'W';
	}
	item = XmStringCreateLtoR(str, XmFONTLIST_DEFAULT_TAG);
	XtVaSetValues(list,
				  XmNitems, &item,
				  XmNitemCount, 1,
				  XtNull);
	XtFree(str);
	XmStringFree(item);
	XtAddCallback(shell, XmNmapCallback, uiFixiateListWidthCallback, list);
}
static void
uiFixiateListWidthCallback(Widget w, XtPointer client, XtPointer call)
{
	Widget form = (Widget)client;
	while (form && !XmIsForm(form))
	{
		form = XtParent(form);
	}
	assert(form != (Widget)0);
	XtRemoveCallback(w, XmNmapCallback, uiFixiateListWidthCallback, client);
	XtVaSetValues(form, XmNresizePolicy, XmRESIZE_NONE, XtNull);
	XmListDeleteAllItems((Widget)client);
}
