/*********************************************************************************************************************************
 * Copyright (c) Des Herriott 1993 - 1994
 *               Erik Kunze   1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: Des Herriott
 *          Erik Kunze
 *
 * $Id: cpu.h,v 1.4 2000/09/24 15:00:55 erik Rel $
 ********************************************************************************************************************************/
#ifndef _XZX_CPU_H
#define _XZX_CPU_H
#ifndef EXTERN_Z80_TYPEDEFS
typedef unsigned char					uns8;
typedef unsigned short					uns16;
typedef unsigned int					uns32;
typedef signed char						sgn8;
typedef signed short					sgn16;
typedef signed int						sgn32;
#endif
typedef void (*NmiCallbackFunc)(void);
typedef union {
#ifdef WORDS_BIGENDIAN
	struct { uns8 h, l; } b;
#else
	struct { uns8 l, h; } b;
#endif
	uns16 w;
} pair;
struct z80 {
	unsigned int state;
	pair af, bc, de, hl;
	pair af_alt,  bc_alt, de_alt, hl_alt;
	pair ix, iy;
	pair sp, pc;
	uns8 r, r7, i;
	uns8 im;
	uns8 iff1, iff2;
	uns8 hidden;
	unsigned int tstates;
	uns8 intVector;
	unsigned int intTstates;
	NmiCallbackFunc nmiCallback;
};
#define CPU_NORMAL						0x0000
#define CPU_HALTED						0x0001
#define CPU_NMI_PENDING					0x0002
#define CPU_INT_PENDING					0x0004
#define CPU_EI_DELAYED					0x0008
#define CPU_DEBUGGED					0x0010
#define AF								Cpu.af.w
#define BC								Cpu.bc.w
#define DE								Cpu.de.w
#define HL								Cpu.hl.w
#define IX								Cpu.ix.w
#define IY								Cpu.iy.w
#define SP								Cpu.sp.w
#define PC								Cpu.pc.w
#define AF1								Cpu.af_alt.w
#define BC1								Cpu.bc_alt.w
#define DE1								Cpu.de_alt.w
#define HL1								Cpu.hl_alt.w
#define A								Cpu.af.b.h
#define F								Cpu.af.b.l
#define B								Cpu.bc.b.h
#define C								Cpu.bc.b.l
#define D								Cpu.de.b.h
#define E								Cpu.de.b.l
#define H								Cpu.hl.b.h
#define L								Cpu.hl.b.l
#define HIX								Cpu.ix.b.h
#define LIX								Cpu.ix.b.l
#define HIY								Cpu.iy.b.h
#define LIY								Cpu.iy.b.l
#define HSP								Cpu.sp.b.h
#define LSP								Cpu.sp.b.l
#define HPC								Cpu.pc.b.h
#define LPC								Cpu.pc.b.l
#define A1								Cpu.af_alt.b.h
#define F1								Cpu.af_alt.b.l
#define B1								Cpu.bc_alt.b.h
#define C1								Cpu.bc_alt.b.l
#define D1								Cpu.de_alt.b.h
#define E1								Cpu.de_alt.b.l
#define H1								Cpu.hl_alt.b.h
#define L1								Cpu.hl_alt.b.l
#define R								Cpu.r
#define R7								Cpu.r7
#define I								Cpu.i
#define IM								Cpu.im
#define IFF1							Cpu.iff1
#define IFF2							Cpu.iff2
#define HIDDEN							Cpu.hidden
#define TSTATES							Cpu.tstates
#define Z80_RDOP						Machine->readOpcode
#define Z80_RDOP2						Machine->readMem
#define Z80_RDMEM						Machine->readMem
#define Z80_WRMEM						Machine->writeMem
#define Z80_IN							Machine->inPort
#define Z80_OUT							Machine->outPort
#define FLAG_C							0x01
#define FLAG_N							0x02
#define FLAG_P							0x04
#define FLAG_V							FLAG_P
#define FLAG_3							0x08
#define FLAG_H							0x10
#define FLAG_5							0x20
#define FLAG_Z							0x40
#define FLAG_S							0x80
#define SET(flag)						(F |= (flag))
#define CLR(flag)						(F &= ~(flag))
#define TST(flag)						((F & (flag)) != 0)
#define CPL(flag)						(F ^= (flag))
#define LOWBYTE(reg)					((reg) & 0xff)
#define HIGHBYTE(reg)					(((reg) >> 8) & 0xff)
#include "z80_macros.h"
extern void Z80_Init(void);
extern void Z80_Reset(void);
extern void Z80_INT(uns8, unsigned int);
extern void Z80_NMI(NmiCallbackFunc);
extern void Z80_DBG(int);
extern unsigned int Z80_Execute(unsigned int);
extern struct z80 Cpu;
#endif
