/*********************************************************************************************************************************
 * Copyright (c) Des Herriott 1993 - 1994
 *               Erik Kunze   1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: Des Herriott
 *          Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char z80_c[] = "$Id: z80.c,v 4.29 2000/12/16 16:33:57 erik Rel $";
#endif
#include <string.h>
#ifdef HAVE_MEMORY_H
# include <memory.h>
#endif
#include "cpu.h"
#include "debug.h"
#include "resource.h"
#include "mem.h"
#include "machine.h"
#include "util.h"
#ifdef XBELL_AUDIO
#include "audio.h"
#endif
#include "breakpoint.h"
#ifdef XZX_IF1
#include "if1.h"
#endif
#include "monitor.h"
#include "snapshot.h"
#include "tape.h"
#ifdef DEBUG
#define DEB(x)							{ if ((CFGGET(debug) & D_INTERRUPT) != 0) { x } }
#else
#define DEB(x)
#endif
#ifndef HAVE_ENOUGH_SWAP
static void z80_DecodeCB(void);
static void z80_DecodeED(void);
static void z80_DecodeDD(void);
static void z80_DecodeFD(void);
#endif
struct z80 Cpu;
static uns8 halfCarryTbl[]    = { 0, 0, FLAG_H, 0, FLAG_H, 0, FLAG_H, FLAG_H };
static uns8 subHalfCarryTbl[] = { 0, FLAG_H, FLAG_H, FLAG_H, 0, 0, 0, FLAG_H };
static uns8 overflowTbl[]     = { 0, FLAG_V, 0, 0, 0, 0, FLAG_V, 0 };
static uns8 subOverflowTbl[]  = { 0, 0, 0, FLAG_V, FLAG_V, 0, 0, 0 };
static uns8 *parityTbl, *sz53pTbl, *sz53Tbl;
static uns8 iniTemp1[4][4] = {
	{ 0, 0, 1, 0 }, { 0, 1, 0, 1 }, { 1, 0, 1, 1 }, { 0, 1, 1, 0 }
};
static uns8 indTemp1[4][4] = {
	{ 0, 1, 0, 0 }, { 1, 0, 0, 1 }, { 0, 0, 1, 0 }, { 0, 1, 0, 1 }
};
static uns8 bregTemp2[256] = {
	0, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1,
	0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0,
	1, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0,
	1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1,
	0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0,
	1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1,
	0, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1,
	0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0,
	1, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0,
	1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1,
	0, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1,
	0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0,
	1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1,
	0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0,
	1, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0,
	1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1
};
void
Z80_Init(void)
{
	int i, j, k, p;
	parityTbl = Malloc(3 * 256, "Z80_Init");
	sz53pTbl = &parityTbl[256];
	sz53Tbl = &parityTbl[512];
	for (i = 0; i < 256; i++)
	{
		for (j = i, k = 8, p = 0; k; k--)
		{
			p ^= j & 1;
			j >>= 1;
		}
		parityTbl[i] = (uns8)(p ? 0 : FLAG_P);
		sz53Tbl[i] = (uns8)(i & (FLAG_S | FLAG_5 | FLAG_3));
		sz53pTbl[i] = sz53Tbl[i] | parityTbl[i];
	}
	sz53Tbl[0] |= FLAG_Z;
	sz53pTbl[0] |= FLAG_Z;
	(void)memset(&Cpu, 0, sizeof(Cpu));
}
void
Z80_Reset(void)
{
	Cpu.state &= ~(CPU_HALTED | CPU_NMI_PENDING | CPU_INT_PENDING | CPU_EI_DELAYED);
	AF = BC = DE = HL = 0;
	AF1 = BC1 = DE1 = HL1 = 0;
	IX = IY = 0xffff;
	SP = PC = 0;
	F = FLAG_Z;
	R = R7 = 0;
	I = IM = IFF1 = IFF2 = 0;
	HIDDEN = 0;
	(void)BpDestroyAll();
}
void
Z80_INT(uns8 ivec, unsigned int itime)
{
	Cpu.state |= CPU_INT_PENDING;
	Cpu.intVector = ivec;
	Cpu.intTstates = itime;
}
void
Z80_NMI(NmiCallbackFunc func)
{
	Cpu.state |= CPU_NMI_PENDING;
	Cpu.nmiCallback = func;
}
void
Z80_DBG(int state)
{
	if (state == 0)
	{
		Cpu.state &= ~CPU_DEBUGGED;
	}
	else
	{
		Cpu.state |= CPU_DEBUGGED;
	}
}
unsigned int
Z80_Execute(unsigned int cycles)
{
	R7 = R & 0x80;
	if ((Cpu.state & (CPU_HALTED | CPU_NMI_PENDING | CPU_INT_PENDING)) == CPU_HALTED)
	{
		unsigned int inc = ((cycles - TSTATES - 1) >> 2) + 1;
		TSTATES += inc << 2;
		R += inc;
		if ((Cpu.state & CPU_EI_DELAYED) != 0)
		{
			Cpu.state &= ~CPU_EI_DELAYED;
		}
	}
	else
	{
		do {
			if (Cpu.state != CPU_NORMAL)
			{
				if ((Cpu.state & CPU_NMI_PENDING) != 0)
				{
					Cpu.state &= ~CPU_NMI_PENDING;
					DEB(Msg(M_DEBUG, "NMI: IM=%d IFF1=%d IFF2=%d", IM, IFF1, IFF2););
					if ((Cpu.state & CPU_HALTED) != 0)
					{
						Cpu.state &= ~CPU_HALTED;
						PC++;
					}
					IFF1 = 0;
					R++;
					TSTATES += 11;
					PUSH(HPC, LPC);
					PC = 0x0066;
					if (Cpu.nmiCallback != (NmiCallbackFunc)0)
					{
						Cpu.nmiCallback();
					}
				}
				if ((Cpu.state & CPU_INT_PENDING) != 0 && TSTATES >= Cpu.intTstates + 32)
				{
					Cpu.state &= ~CPU_INT_PENDING;
				}
				if ((Cpu.state & (CPU_INT_PENDING | CPU_EI_DELAYED)) == CPU_INT_PENDING)
				{
					DEB(Msg(M_DEBUG, "INT: IM=%d IFF1=%d IFF2=%d", IM, IFF1, IFF2););
					if (IFF1 != 0)
					{
						if ((Cpu.state & CPU_HALTED) != 0)
						{
							Cpu.state &= ~CPU_HALTED;
							PC++;
						}
						IFF1 = IFF2 = 0;
						R++;
						switch (IM)
						{
							case 0:
								TSTATES += 13;
								PUSH(HPC, LPC);
								PC = 0x0038;
								break;
							case 1:
								TSTATES += 13;
								PUSH(HPC, LPC);
								PC = 0x0038;
								break;
							case 2:
								TSTATES += 19;
								PUSH(HPC, LPC);
								{
									uns16 t16 = (I << 8) | Cpu.intVector;
									LPC = Z80_RDMEM(t16++);
									HPC = Z80_RDMEM(t16);
								}
								break;
						}
					}
#ifdef DEBUG_HALT
					else if ((Cpu.state & CPU_HALTED) != 0)
					{
						Cpu.state &= ~CPU_HALTED;
						Msg(M_WARN, "leaving HALT mode (forced)");
					}
#endif
				}
				Cpu.state &= ~CPU_EI_DELAYED;
				if ((Cpu.state & CPU_DEBUGGED) != 0)
				{
					MonitorEnter(MONITOR_SINGLESTEP);
				}
			}
#ifdef DEBUG
			if ((CFGGET(debug) & (D_CPU | D_SYSVAR | D_MNEMO)) != 0)
			{
				DebugZ80();
			}
#endif
			{
				uns8 op = Z80_RDOP(PC++);
				R++;
				switch (op)
				{
#					include "z80_ops.c"
				}
			}
		} while (TSTATES < cycles);
	}
	R = (R & 0x7f) | R7;
	return (TSTATES - cycles);
}
#ifndef HAVE_ENOUGH_SWAP
static void
z80_DecodeCB(void)
{
	uns8 op = Z80_RDOP2(PC++);
	R++;
	switch(op)
	{
#		include "z80_cbops.c"
	}
}
static void
z80_DecodeED(void)
{
	uns8 op = Z80_RDOP2(PC++);
	R++;
	switch(op)
	{
#		include "z80_edops.c"
	}
}
static void
z80_DecodeDD(void)
{
	uns8 op = Z80_RDOP2(PC++);
	R++;
	switch(op)
	{
#		define XX  IX
#		define LXX LIX
#		define HXX HIX
#		include "z80_xxops.c"
#		undef XX
#		undef LXX
#		undef HXX
	}
}
static void
z80_DecodeFD(void)
{
	uns8 op = Z80_RDOP2(PC++);
	R++;
	switch(op)
	{
#		define NO_RCSID
#		define XX  IY
#		define LXX LIY
#		define HXX HIY
#		include "z80_xxops.c"
#		undef XX
#		undef LXX
#		undef HXX
#		undef NO_RCSID
	}
}
#endif
